from datetime import timedelta, datetime, time, tzinfo
from enum import IntEnum, unique
from typing import List, Any, Dict, Tuple, NamedTuple
from data_aggregator_sdk.constants.enums import IntegrationV0MessageEvent
from data_aggregator_sdk.integration_message import IntegrationV0MessageData

from data_gateway_sdk.utils.buf_ref import BufRef
from data_gateway_sdk.utils.days_ago_calculation import days_ago_calculation
from data_gateway_sdk.utils.packet import Packet


# PACKET (128 bits)   smpm_ul_device_energy_16b_journal
#
# RESULT int:        952893555
# RESULT bin:  MSB   00000000 00000000 00000000 00000000 00000000 00000000 00000000 00000000 00000000 00000000 00000000 00000000 00111000 11001100 00000000 01110011   LSB                                      # noqa: E501
# RESULT hex:  LE    73 00 cc 38 00 00 00 00 00 00 00 00 00 00 00 00
#
# name                    type                              size  value(int)                                                                                                                        data(bits)  # noqa: E501
# ----------------------  --------------------------------  ----  ----------  --------------------------------------------------------------------------------------------------------------------------------  # noqa: E501
# packet_type_id.0.VAL    u7                                   7         115                                                                                                                           1110011  # noqa: E501
# packet_type_id.0.DFF    bool                                 1           0                                                                                                                          0         # noqa: E501
# days_ago                timedelta                            7           0                                                                                                                   0000000          # noqa: E501
# valid                   bool                                 1           0                                                                                                                  0                 # noqa: E501
# time_offset             timedelta                            6          12                                                                                                            001100                  # noqa: E501
# journal.0.event.offset  timedelta                            5           3                                                                                                       00011                        # noqa: E501
# journal.0.event.code    SmpmUlDeviceEnergy16bJournalType     8           7                                                                                               00000111                             # noqa: E501
# journal.1.event.offset  timedelta                            5           0                                                                                          00000                                     # noqa: E501
# journal.1.event.code    SmpmUlDeviceEnergy16bJournalType     8           0                                                                                  00000000                                          # noqa: E501
# journal.2.event.offset  timedelta                            5           0                                                                             00000                                                  # noqa: E501
# journal.2.event.code    SmpmUlDeviceEnergy16bJournalType     8           0                                                                     00000000                                                       # noqa: E501
# journal.3.event.offset  timedelta                            5           0                                                                00000                                                               # noqa: E501
# journal.3.event.code    SmpmUlDeviceEnergy16bJournalType     8           0                                                        00000000                                                                    # noqa: E501
# journal.4.event.offset  timedelta                            5           0                                                   00000                                                                            # noqa: E501
# journal.4.event.code    SmpmUlDeviceEnergy16bJournalType     8           0                                           00000000                                                                                 # noqa: E501
# journal.5.event.offset  timedelta                            5           0                                      00000
# journal.5.event.code    SmpmUlDeviceEnergy16bJournalType     8           0                              00000000
# journal.6.event.offset  timedelta                            5           0                         00000
# journal.6.event.code    SmpmUlDeviceEnergy16bJournalType     8           0                 00000000
# journal.7.event.offset  timedelta                            5           0            00000
# journal.7.event.code    SmpmUlDeviceEnergy16bJournalType     8           0    00000000
# RESERVED                u2                                   2           0  00


@unique
class SmpmUlDeviceEnergy16bJournalType(IntEnum):
    NONE = 0
    SUCCESSFUL_AUTO_DIAGNOSTIC = 1
    SWITCH_WINTER_DAYLIGHT = 2
    SWITCH_SUMMER_DAYLIGHT = 3
    SETUP_UPDATE = 5
    RECORD_DATETIME = 6
    CHANGE_OFFSET_DAILY_CLOCK = 7
    PERMISSION_SWITCH_DAYLIGHT_ON = 8
    PERMISSION_SWITCH_DAYLIGHT_OFF = 9
    CHANGE_DATE_TIME_SWITCH_DAYLIGHT = 10
    ERASE_EEPROM = 12
    NULLIFY_TARIFF_ACCUMULATION = 13
    NULLIFY_INTERVAL_ACCUMULATION = 14
    RESET_PASSWORD = 15
    RESET_POWER_LOST_TIME_COUNTER = 16
    RESET_MAGNET_IMPACT_TIME_COUNTER = 17
    RESET_POWER_INCREASE_TIME_COUNTER = 18
    RESET_POWER_DECREASE_TIME_COUNTER = 19
    RESET_MAINTS_FREQ_DIVERGENCE_TIME_COUNTER = 20
    RESET_POWER_OVER_LIMIT_TIME_COUNTER = 22
    CHANGE_CAPACITY_DATA_LCD = 25
    CHANGE_TARIFF_METHODS = 26
    CHANGE_TARIFF_PROGRAMS = 27
    CHANGE_ACTUAL_SEASON_SCHEDULES = 28
    CHANGE_CONSUMPTION_LIMIT = 34
    CHANGE_LOW_THRESHOLD_VOLTAGE = 35
    CHANGE_HIGH_THRESHOLD_VOLTAGE = 36
    CHANGE_MAINTS_FREQ_THRESHOLD = 37
    CHANGE_THRESHOLD_LOW_CONSUMPTION = 39
    RECHARGE_ENERGY_PAYMENT = 40
    UNSUCCESSFUL_AUTO_DIAGNOSTIC_INTERNAL_CLOCK = 58
    ABNORMAL_COUNTER_AUTOSTART = 59
    EXTERNAL_POWER_LOST = 60
    EXTERNAL_POWER_DETECTED = 61
    START_POWER_OVER_LIMIT = 68
    STOP_POWER_OVER_LIMIT = 69
    ENERGY_OVER_LIMIT_1 = 70
    ENERGY_OVER_LIMIT_2 = 71
    ENERGY_OVER_LIMIT_3 = 72
    WRONG_PASSWORD_BLOCK = 73
    WRONG_PASSWORD_APPEAL = 74
    EXHAUST_DAILY_BATTERY_LIFE_LIMIT = 75
    START_MAGNET_IMPACT = 76
    STOP_MAGNET_IMPACT = 77
    VIOLATION_TERMINAL_BLOCK_SEAL = 78
    RECOVERY_TERMINAL_BLOCK_SEAL = 79
    VIOLATION_CASE_SEAL = 80
    RECOVERY_CASE_SEAL = 81
    TIME_OUT_SYNC_LIMIT = 84
    CRITICAL_DIVERGENCE_TIME = 85
    OVERHEAT_COUNTER_START = 90
    OVERHEAT_COUNTER_STOP = 91
    UNSUCCESSFUL_AUTO_DIAGNOSTIC_MEMORY = 92
    LOW_BATTERY_CAPACITY = 94
    RECOVERY_BATTERY_WORKING_VOLTAGE = 95
    LOW_CONSUMPTION = 96
    RESET_FLAG_LOW_CONSUMPTION = 97
    CHANGE_VALIDATION_SETTINGS = 113
    UNSUCCESSFUL_AUTO_DIAGNOSTIC_MEASUREMENT_BLOCK = 116
    UNSUCCESSFUL_AUTO_DIAGNOSTIC_CALCULATION_BLOCK = 117
    UNSUCCESSFUL_AUTO_DIAGNOSTIC_POWER_BLOCK = 118
    UNSUCCESSFUL_AUTO_DIAGNOSTIC_SCREEN = 119
    UNSUCCESSFUL_AUTO_DIAGNOSTIC_RADIO = 120
    MAINS_VOLTAGE_LOST_PHASE_A_START = 134
    MAINS_VOLTAGE_LOST_PHASE_A_STOP = 135
    MAINS_VOLTAGE_LOST_PHASE_B_START = 136
    MAINS_VOLTAGE_LOST_PHASE_B_STOP = 137
    MAINS_VOLTAGE_LOST_PHASE_C_START = 138
    MAINS_VOLTAGE_LOST_PHASE_C_STOP = 139
    VOLTAGE_LAYDOWN_PHASE_A_START = 140
    VOLTAGE_LAYDOWN_PHASE_A_STOP = 141
    VOLTAGE_LAYDOWN_PHASE_B_START = 142
    VOLTAGE_LAYDOWN_PHASE_B_STOP = 143
    VOLTAGE_LAYDOWN_PHASE_C_START = 144
    VOLTAGE_LAYDOWN_PHASE_C_STOP = 145
    OVERVOLTAGE_PHASE_A_START = 146
    OVERVOLTAGE_PHASE_A_STOP = 147
    OVERVOLTAGE_PHASE_B_START = 148
    OVERVOLTAGE_PHASE_B_STOP = 149
    OVERVOLTAGE_PHASE_C_START = 150
    OVERVOLTAGE_PHASE_C_STOP = 151
    OVERCURRENT_PHASE_A_START = 152
    OVERCURRENT_PHASE_A_STOP = 153
    OVERCURRENT_PHASE_B_START = 154
    OVERCURRENT_PHASE_B_STOP = 155
    OVERCURRENT_PHASE_C_START = 156
    OVERCURRENT_PHASE_C_STOP = 157
    CURRENT_SUM_THRESHOLD_LOW_START = 158
    CURRENT_SUM_THRESHOLD_LOW_STOP = 159
    FREQ_OUT_PHASE_A_START = 160
    FREQ_OUT_PHASE_A_STOP = 161
    FREQ_OUT_PHASE_B_START = 162
    FREQ_OUT_PHASE_B_STOP = 163
    FREQ_OUT_PHASE_C_START = 164
    FREQ_OUT_PHASE_C_STOP = 165
    PHASE_ORDER_DISTURBANCE_START = 166
    PHASE_ORDER_DISTURBANCE_STOP = 167
    RADIO_IMPACT_START = 169
    RADIO_IMPACT_STOP = 170
    DAYLIGHT_TIME_SWITCH = 173
    DAYLIGHT_TIME_MODE_DATES_CHANGE = 174
    INTERNAL_CLOCK_SYNC = 175
    METROLOGY_CHANGE = 176
    PROFILE_CONF_CHANGE = 177
    TARIFFICATION_METHOD_CHANGE = 178
    PERMISSION_CHANGE_SETTINGS_POWER_CONTROL = 179
    CONTROL_LEVEL_MAINS_CHANGE = 180
    PERMISSION_CHANGE_SETTINGS_CONSUMPTION_CONTROL = 181
    LOAD_RELAY_CONDITION_SETTINGS_CHANGE = 182
    SIGNALIZATION_RELAY_CONDITION_SETTINGS_CHANGE = 183
    INTERFACE_SIGNALIZATION_CONDITION_SETTINGS_CHANGE = 184
    INDICATION_SETTINGS_CHANGE = 185
    SOUND_SIGNAL_CONDITION_SETTINGS_CHANGE = 186
    LOAD_RELAY_STATE_CHANGE = 187
    SIGNALIZATION_RELAY_STATE_CHANGE = 188

    def __repr__(self) -> str:
        return f'{type(self).__name__}.{self.name}'


EVENTS_MAP = {
    SmpmUlDeviceEnergy16bJournalType.NONE: IntegrationV0MessageEvent.NONE,
    SmpmUlDeviceEnergy16bJournalType.SUCCESSFUL_AUTO_DIAGNOSTIC: IntegrationV0MessageEvent.SUCCESSFUL_AUTO_DIAGNOSTIC,
    SmpmUlDeviceEnergy16bJournalType.SETUP_UPDATE: IntegrationV0MessageEvent.SETUP_UPDATE,
    SmpmUlDeviceEnergy16bJournalType.SWITCH_WINTER_DAYLIGHT: IntegrationV0MessageEvent.SWITCH_WINTER_DAYLIGHT,
    SmpmUlDeviceEnergy16bJournalType.SWITCH_SUMMER_DAYLIGHT: IntegrationV0MessageEvent.SWITCH_SUMMER_DAYLIGHT,
    SmpmUlDeviceEnergy16bJournalType.RECORD_DATETIME: IntegrationV0MessageEvent.RECORD_DATETIME,
    SmpmUlDeviceEnergy16bJournalType.CHANGE_OFFSET_DAILY_CLOCK: IntegrationV0MessageEvent.CHANGE_OFFSET_DAILY_CLOCK,
    SmpmUlDeviceEnergy16bJournalType.PERMISSION_SWITCH_DAYLIGHT_ON: IntegrationV0MessageEvent.PERMISSION_SWITCH_DAYLIGHT_ON,
    SmpmUlDeviceEnergy16bJournalType.PERMISSION_SWITCH_DAYLIGHT_OFF: IntegrationV0MessageEvent.PERMISSION_SWITCH_DAYLIGHT_OFF,
    SmpmUlDeviceEnergy16bJournalType.CHANGE_DATE_TIME_SWITCH_DAYLIGHT: IntegrationV0MessageEvent.CHANGE_DATE_TIME_SWITCH_DAYLIGHT,
    SmpmUlDeviceEnergy16bJournalType.ERASE_EEPROM: IntegrationV0MessageEvent.ERASE_EEPROM,
    SmpmUlDeviceEnergy16bJournalType.NULLIFY_TARIFF_ACCUMULATION: IntegrationV0MessageEvent.NULLIFY_TARIFF_ACCUMULATION,
    SmpmUlDeviceEnergy16bJournalType.NULLIFY_INTERVAL_ACCUMULATION: IntegrationV0MessageEvent.NULLIFY_INTERVAL_ACCUMULATION,
    SmpmUlDeviceEnergy16bJournalType.RESET_PASSWORD: IntegrationV0MessageEvent.RESET_PASSWORD,
    SmpmUlDeviceEnergy16bJournalType.RESET_POWER_LOST_TIME_COUNTER: IntegrationV0MessageEvent.RESET_POWER_LOST_TIME_COUNTER,
    SmpmUlDeviceEnergy16bJournalType.RESET_MAGNET_IMPACT_TIME_COUNTER: IntegrationV0MessageEvent.RESET_MAGNET_IMPACT_TIME_COUNTER,
    SmpmUlDeviceEnergy16bJournalType.RESET_POWER_INCREASE_TIME_COUNTER: IntegrationV0MessageEvent.RESET_POWER_INCREASE_TIME_COUNTER,
    SmpmUlDeviceEnergy16bJournalType.RESET_POWER_DECREASE_TIME_COUNTER: IntegrationV0MessageEvent.RESET_POWER_DECREASE_TIME_COUNTER,
    SmpmUlDeviceEnergy16bJournalType.RESET_MAINTS_FREQ_DIVERGENCE_TIME_COUNTER: IntegrationV0MessageEvent.RESET_MAINTS_FREQ_DIVERGENCE_TIME_COUNTER,
    SmpmUlDeviceEnergy16bJournalType.RESET_POWER_OVER_LIMIT_TIME_COUNTER: IntegrationV0MessageEvent.RESET_POWER_OVER_LIMIT_TIME_COUNTER,
    SmpmUlDeviceEnergy16bJournalType.CHANGE_CAPACITY_DATA_LCD: IntegrationV0MessageEvent.CHANGE_CAPACITY_DATA_LCD,
    SmpmUlDeviceEnergy16bJournalType.CHANGE_TARIFF_METHODS: IntegrationV0MessageEvent.CHANGE_TARIFF_METHODS,
    SmpmUlDeviceEnergy16bJournalType.CHANGE_TARIFF_PROGRAMS: IntegrationV0MessageEvent.CHANGE_TARIFF_PROGRAMS,
    SmpmUlDeviceEnergy16bJournalType.CHANGE_ACTUAL_SEASON_SCHEDULES: IntegrationV0MessageEvent.CHANGE_ACTUAL_SEASON_SCHEDULES,
    SmpmUlDeviceEnergy16bJournalType.CHANGE_CONSUMPTION_LIMIT: IntegrationV0MessageEvent.CHANGE_CONSUMPTION_LIMIT,
    SmpmUlDeviceEnergy16bJournalType.CHANGE_LOW_THRESHOLD_VOLTAGE: IntegrationV0MessageEvent.CHANGE_LOW_THRESHOLD_VOLTAGE,
    SmpmUlDeviceEnergy16bJournalType.CHANGE_HIGH_THRESHOLD_VOLTAGE: IntegrationV0MessageEvent.CHANGE_HIGH_THRESHOLD_VOLTAGE,
    SmpmUlDeviceEnergy16bJournalType.CHANGE_MAINTS_FREQ_THRESHOLD: IntegrationV0MessageEvent.CHANGE_MAINTS_FREQ_THRESHOLD,
    SmpmUlDeviceEnergy16bJournalType.CHANGE_THRESHOLD_LOW_CONSUMPTION: IntegrationV0MessageEvent.CHANGE_THRESHOLD_LOW_CONSUMPTION,
    SmpmUlDeviceEnergy16bJournalType.RECHARGE_ENERGY_PAYMENT: IntegrationV0MessageEvent.RECHARGE_ENERGY_PAYMENT,
    SmpmUlDeviceEnergy16bJournalType.UNSUCCESSFUL_AUTO_DIAGNOSTIC_INTERNAL_CLOCK: IntegrationV0MessageEvent.UNSUCCESSFUL_AUTO_DIAGNOSTIC_INTERNAL_CLOCK,
    SmpmUlDeviceEnergy16bJournalType.ABNORMAL_COUNTER_AUTOSTART: IntegrationV0MessageEvent.ABNORMAL_COUNTER_AUTOSTART,
    SmpmUlDeviceEnergy16bJournalType.EXTERNAL_POWER_LOST: IntegrationV0MessageEvent.EXTERNAL_POWER_LOST,
    SmpmUlDeviceEnergy16bJournalType.EXTERNAL_POWER_DETECTED: IntegrationV0MessageEvent.EXTERNAL_POWER_DETECTED,
    SmpmUlDeviceEnergy16bJournalType.START_POWER_OVER_LIMIT: IntegrationV0MessageEvent.START_POWER_OVER_LIMIT,
    SmpmUlDeviceEnergy16bJournalType.STOP_POWER_OVER_LIMIT: IntegrationV0MessageEvent.STOP_POWER_OVER_LIMIT,
    SmpmUlDeviceEnergy16bJournalType.ENERGY_OVER_LIMIT_1: IntegrationV0MessageEvent.ENERGY_OVER_LIMIT_1,
    SmpmUlDeviceEnergy16bJournalType.ENERGY_OVER_LIMIT_2: IntegrationV0MessageEvent.ENERGY_OVER_LIMIT_2,
    SmpmUlDeviceEnergy16bJournalType.ENERGY_OVER_LIMIT_3: IntegrationV0MessageEvent.ENERGY_OVER_LIMIT_3,
    SmpmUlDeviceEnergy16bJournalType.WRONG_PASSWORD_BLOCK: IntegrationV0MessageEvent.WRONG_PASSWORD_BLOCK,
    SmpmUlDeviceEnergy16bJournalType.WRONG_PASSWORD_APPEAL: IntegrationV0MessageEvent.WRONG_PASSWORD_APPEAL,
    SmpmUlDeviceEnergy16bJournalType.EXHAUST_DAILY_BATTERY_LIFE_LIMIT: IntegrationV0MessageEvent.EXHAUST_DAILY_BATTERY_LIFE_LIMIT,
    SmpmUlDeviceEnergy16bJournalType.START_MAGNET_IMPACT: IntegrationV0MessageEvent.START_MAGNET_IMPACT,
    SmpmUlDeviceEnergy16bJournalType.STOP_MAGNET_IMPACT: IntegrationV0MessageEvent.STOP_MAGNET_IMPACT,
    SmpmUlDeviceEnergy16bJournalType.VIOLATION_TERMINAL_BLOCK_SEAL: IntegrationV0MessageEvent.VIOLATION_TERMINAL_BLOCK_SEAL,
    SmpmUlDeviceEnergy16bJournalType.RECOVERY_TERMINAL_BLOCK_SEAL: IntegrationV0MessageEvent.RECOVERY_TERMINAL_BLOCK_SEAL,
    SmpmUlDeviceEnergy16bJournalType.VIOLATION_CASE_SEAL: IntegrationV0MessageEvent.VIOLATION_CASE_SEAL,
    SmpmUlDeviceEnergy16bJournalType.RECOVERY_CASE_SEAL: IntegrationV0MessageEvent.RECOVERY_CASE_SEAL,
    SmpmUlDeviceEnergy16bJournalType.TIME_OUT_SYNC_LIMIT: IntegrationV0MessageEvent.TIME_OUT_SYNC_LIMIT,
    SmpmUlDeviceEnergy16bJournalType.CRITICAL_DIVERGENCE_TIME: IntegrationV0MessageEvent.CRITICAL_DIVERGENCE_TIME,
    SmpmUlDeviceEnergy16bJournalType.OVERHEAT_COUNTER_START: IntegrationV0MessageEvent.OVERHEAT_COUNTER_START,
    SmpmUlDeviceEnergy16bJournalType.OVERHEAT_COUNTER_STOP: IntegrationV0MessageEvent.OVERHEAT_COUNTER_STOP,
    SmpmUlDeviceEnergy16bJournalType.UNSUCCESSFUL_AUTO_DIAGNOSTIC_MEMORY: IntegrationV0MessageEvent.UNSUCCESSFUL_AUTO_DIAGNOSTIC_MEMORY,
    SmpmUlDeviceEnergy16bJournalType.LOW_BATTERY_CAPACITY: IntegrationV0MessageEvent.LOW_BATTERY_CAPACITY,
    SmpmUlDeviceEnergy16bJournalType.RECOVERY_BATTERY_WORKING_VOLTAGE: IntegrationV0MessageEvent.RECOVERY_BATTERY_WORKING_VOLTAGE,
    SmpmUlDeviceEnergy16bJournalType.LOW_CONSUMPTION: IntegrationV0MessageEvent.LOW_CONSUMPTION,
    SmpmUlDeviceEnergy16bJournalType.RESET_FLAG_LOW_CONSUMPTION: IntegrationV0MessageEvent.RESET_FLAG_LOW_CONSUMPTION,
    SmpmUlDeviceEnergy16bJournalType.CHANGE_VALIDATION_SETTINGS: IntegrationV0MessageEvent.CHANGE_VALIDATION_SETTINGS,
    SmpmUlDeviceEnergy16bJournalType.UNSUCCESSFUL_AUTO_DIAGNOSTIC_MEASUREMENT_BLOCK: IntegrationV0MessageEvent.UNSUCCESSFUL_AUTO_DIAGNOSTIC_MEASUREMENT_BLOCK,
    SmpmUlDeviceEnergy16bJournalType.UNSUCCESSFUL_AUTO_DIAGNOSTIC_CALCULATION_BLOCK: IntegrationV0MessageEvent.UNSUCCESSFUL_AUTO_DIAGNOSTIC_CALCULATION_BLOCK,
    SmpmUlDeviceEnergy16bJournalType.UNSUCCESSFUL_AUTO_DIAGNOSTIC_POWER_BLOCK: IntegrationV0MessageEvent.UNSUCCESSFUL_AUTO_DIAGNOSTIC_POWER_BLOCK,
    SmpmUlDeviceEnergy16bJournalType.UNSUCCESSFUL_AUTO_DIAGNOSTIC_SCREEN: IntegrationV0MessageEvent.UNSUCCESSFUL_AUTO_DIAGNOSTIC_SCREEN,
    SmpmUlDeviceEnergy16bJournalType.UNSUCCESSFUL_AUTO_DIAGNOSTIC_RADIO: IntegrationV0MessageEvent.UNSUCCESSFUL_AUTO_DIAGNOSTIC_RADIO,
    SmpmUlDeviceEnergy16bJournalType.MAINS_VOLTAGE_LOST_PHASE_A_START: IntegrationV0MessageEvent.MAINS_VOLTAGE_LOST_PHASE_A_START,
    SmpmUlDeviceEnergy16bJournalType.MAINS_VOLTAGE_LOST_PHASE_A_STOP: IntegrationV0MessageEvent.MAINS_VOLTAGE_LOST_PHASE_A_STOP,
    SmpmUlDeviceEnergy16bJournalType.MAINS_VOLTAGE_LOST_PHASE_B_START: IntegrationV0MessageEvent.MAINS_VOLTAGE_LOST_PHASE_B_START,
    SmpmUlDeviceEnergy16bJournalType.MAINS_VOLTAGE_LOST_PHASE_B_STOP: IntegrationV0MessageEvent.MAINS_VOLTAGE_LOST_PHASE_B_STOP,
    SmpmUlDeviceEnergy16bJournalType.MAINS_VOLTAGE_LOST_PHASE_C_START: IntegrationV0MessageEvent.MAINS_VOLTAGE_LOST_PHASE_C_START,
    SmpmUlDeviceEnergy16bJournalType.MAINS_VOLTAGE_LOST_PHASE_C_STOP: IntegrationV0MessageEvent.MAINS_VOLTAGE_LOST_PHASE_C_STOP,
    SmpmUlDeviceEnergy16bJournalType.VOLTAGE_LAYDOWN_PHASE_A_START: IntegrationV0MessageEvent.VOLTAGE_LAYDOWN_PHASE_A_START,
    SmpmUlDeviceEnergy16bJournalType.VOLTAGE_LAYDOWN_PHASE_A_STOP: IntegrationV0MessageEvent.VOLTAGE_LAYDOWN_PHASE_A_STOP,
    SmpmUlDeviceEnergy16bJournalType.VOLTAGE_LAYDOWN_PHASE_B_START: IntegrationV0MessageEvent.VOLTAGE_LAYDOWN_PHASE_B_START,
    SmpmUlDeviceEnergy16bJournalType.VOLTAGE_LAYDOWN_PHASE_B_STOP: IntegrationV0MessageEvent.VOLTAGE_LAYDOWN_PHASE_B_STOP,
    SmpmUlDeviceEnergy16bJournalType.VOLTAGE_LAYDOWN_PHASE_C_START: IntegrationV0MessageEvent.VOLTAGE_LAYDOWN_PHASE_C_START,
    SmpmUlDeviceEnergy16bJournalType.VOLTAGE_LAYDOWN_PHASE_C_STOP: IntegrationV0MessageEvent.VOLTAGE_LAYDOWN_PHASE_C_STOP,
    SmpmUlDeviceEnergy16bJournalType.OVERVOLTAGE_PHASE_A_START: IntegrationV0MessageEvent.OVERVOLTAGE_PHASE_A_START,
    SmpmUlDeviceEnergy16bJournalType.OVERVOLTAGE_PHASE_A_STOP: IntegrationV0MessageEvent.OVERVOLTAGE_PHASE_A_STOP,
    SmpmUlDeviceEnergy16bJournalType.OVERVOLTAGE_PHASE_B_START: IntegrationV0MessageEvent.OVERVOLTAGE_PHASE_B_START,
    SmpmUlDeviceEnergy16bJournalType.OVERVOLTAGE_PHASE_B_STOP: IntegrationV0MessageEvent.OVERVOLTAGE_PHASE_B_STOP,
    SmpmUlDeviceEnergy16bJournalType.OVERVOLTAGE_PHASE_C_START: IntegrationV0MessageEvent.OVERVOLTAGE_PHASE_C_START,
    SmpmUlDeviceEnergy16bJournalType.OVERVOLTAGE_PHASE_C_STOP: IntegrationV0MessageEvent.OVERVOLTAGE_PHASE_C_STOP,
    SmpmUlDeviceEnergy16bJournalType.OVERCURRENT_PHASE_A_START: IntegrationV0MessageEvent.OVERCURRENT_PHASE_A_START,
    SmpmUlDeviceEnergy16bJournalType.OVERCURRENT_PHASE_A_STOP: IntegrationV0MessageEvent.OVERCURRENT_PHASE_A_STOP,
    SmpmUlDeviceEnergy16bJournalType.OVERCURRENT_PHASE_B_START: IntegrationV0MessageEvent.OVERCURRENT_PHASE_B_START,
    SmpmUlDeviceEnergy16bJournalType.OVERCURRENT_PHASE_B_STOP: IntegrationV0MessageEvent.OVERCURRENT_PHASE_B_STOP,
    SmpmUlDeviceEnergy16bJournalType.OVERCURRENT_PHASE_C_START: IntegrationV0MessageEvent.OVERCURRENT_PHASE_C_START,
    SmpmUlDeviceEnergy16bJournalType.OVERCURRENT_PHASE_C_STOP: IntegrationV0MessageEvent.OVERCURRENT_PHASE_C_STOP,
    SmpmUlDeviceEnergy16bJournalType.CURRENT_SUM_THRESHOLD_LOW_START: IntegrationV0MessageEvent.CURRENT_SUM_THRESHOLD_LOW_START,
    SmpmUlDeviceEnergy16bJournalType.CURRENT_SUM_THRESHOLD_LOW_STOP: IntegrationV0MessageEvent.CURRENT_SUM_THRESHOLD_LOW_STOP,
    SmpmUlDeviceEnergy16bJournalType.FREQ_OUT_PHASE_A_START: IntegrationV0MessageEvent.FREQ_OUT_PHASE_A_START,
    SmpmUlDeviceEnergy16bJournalType.FREQ_OUT_PHASE_A_STOP: IntegrationV0MessageEvent.FREQ_OUT_PHASE_A_STOP,
    SmpmUlDeviceEnergy16bJournalType.FREQ_OUT_PHASE_B_START: IntegrationV0MessageEvent.FREQ_OUT_PHASE_B_START,
    SmpmUlDeviceEnergy16bJournalType.FREQ_OUT_PHASE_B_STOP: IntegrationV0MessageEvent.FREQ_OUT_PHASE_B_STOP,
    SmpmUlDeviceEnergy16bJournalType.FREQ_OUT_PHASE_C_START: IntegrationV0MessageEvent.FREQ_OUT_PHASE_C_START,
    SmpmUlDeviceEnergy16bJournalType.FREQ_OUT_PHASE_C_STOP: IntegrationV0MessageEvent.FREQ_OUT_PHASE_C_STOP,
    SmpmUlDeviceEnergy16bJournalType.PHASE_ORDER_DISTURBANCE_START: IntegrationV0MessageEvent.PHASE_ORDER_DISTURBANCE_START,
    SmpmUlDeviceEnergy16bJournalType.PHASE_ORDER_DISTURBANCE_STOP: IntegrationV0MessageEvent.PHASE_ORDER_DISTURBANCE_STOP,
    SmpmUlDeviceEnergy16bJournalType.RADIO_IMPACT_START: IntegrationV0MessageEvent.RADIO_IMPACT_START,
    SmpmUlDeviceEnergy16bJournalType.RADIO_IMPACT_STOP: IntegrationV0MessageEvent.RADIO_IMPACT_STOP,
    SmpmUlDeviceEnergy16bJournalType.DAYLIGHT_TIME_SWITCH: IntegrationV0MessageEvent.DAYLIGHT_TIME_SWITCH,
    SmpmUlDeviceEnergy16bJournalType.DAYLIGHT_TIME_MODE_DATES_CHANGE: IntegrationV0MessageEvent.DAYLIGHT_TIME_MODE_DATES_CHANGE,
    SmpmUlDeviceEnergy16bJournalType.INTERNAL_CLOCK_SYNC: IntegrationV0MessageEvent.INTERNAL_CLOCK_SYNC,
    SmpmUlDeviceEnergy16bJournalType.METROLOGY_CHANGE: IntegrationV0MessageEvent.METROLOGY_CHANGE,
    SmpmUlDeviceEnergy16bJournalType.PROFILE_CONF_CHANGE: IntegrationV0MessageEvent.PROFILE_CONF_CHANGE,
    SmpmUlDeviceEnergy16bJournalType.TARIFFICATION_METHOD_CHANGE: IntegrationV0MessageEvent.TARIFFICATION_METHOD_CHANGE,
    SmpmUlDeviceEnergy16bJournalType.PERMISSION_CHANGE_SETTINGS_POWER_CONTROL: IntegrationV0MessageEvent.PERMISSION_CHANGE_SETTINGS_POWER_CONTROL,
    SmpmUlDeviceEnergy16bJournalType.CONTROL_LEVEL_MAINS_CHANGE: IntegrationV0MessageEvent.CONTROL_LEVEL_MAINS_CHANGE,
    SmpmUlDeviceEnergy16bJournalType.PERMISSION_CHANGE_SETTINGS_CONSUMPTION_CONTROL: IntegrationV0MessageEvent.PERMISSION_CHANGE_SETTINGS_CONSUMPTION_CONTROL,
    SmpmUlDeviceEnergy16bJournalType.LOAD_RELAY_CONDITION_SETTINGS_CHANGE: IntegrationV0MessageEvent.LOAD_RELAY_CONDITION_SETTINGS_CHANGE,
    SmpmUlDeviceEnergy16bJournalType.SIGNALIZATION_RELAY_CONDITION_SETTINGS_CHANGE: IntegrationV0MessageEvent.SIGNALIZATION_RELAY_CONDITION_SETTINGS_CHANGE,
    SmpmUlDeviceEnergy16bJournalType.INTERFACE_SIGNALIZATION_CONDITION_SETTINGS_CHANGE: IntegrationV0MessageEvent.INTERFACE_SIGNALIZATION_CONDITION_SETTINGS_CHANGE,
    SmpmUlDeviceEnergy16bJournalType.INDICATION_SETTINGS_CHANGE: IntegrationV0MessageEvent.INDICATION_SETTINGS_CHANGE,
    SmpmUlDeviceEnergy16bJournalType.SOUND_SIGNAL_CONDITION_SETTINGS_CHANGE: IntegrationV0MessageEvent.SOUND_SIGNAL_CONDITION_SETTINGS_CHANGE,
    SmpmUlDeviceEnergy16bJournalType.LOAD_RELAY_STATE_CHANGE: IntegrationV0MessageEvent.LOAD_RELAY_STATE_CHANGE,
    SmpmUlDeviceEnergy16bJournalType.SIGNALIZATION_RELAY_STATE_CHANGE: IntegrationV0MessageEvent.SIGNALIZATION_RELAY_STATE_CHANGE,
}
assert len(set(SmpmUlDeviceEnergy16bJournalType)) == len(set(EVENTS_MAP.keys())), "Wrong length of events codes"


class EventData(NamedTuple):
    offset: timedelta
    code: SmpmUlDeviceEnergy16bJournalType


class SmpmUlDeviceEnergy16BJournalData(Packet):
    days_ago: timedelta
    valid: bool
    time_offset: timedelta
    journal: Tuple[EventData, EventData, EventData, EventData, EventData, EventData, EventData, EventData]

    def serialize(self) -> bytes:
        data = self
        result = 0
        size = 0
        result |= ((115) & (2 ** (7) - 1)) << size
        size += 7
        result |= ((0) & (2 ** (1) - 1)) << size
        size += 1
        isinstance(data.days_ago, (int, timedelta))
        days_ago_tmp1 = int(data.days_ago.total_seconds() // 86400 if isinstance(data.days_ago, timedelta) else data.days_ago // 86400)
        assert 0 <= days_ago_tmp1 <= 127
        result |= ((days_ago_tmp1) & (2 ** (7) - 1)) << size
        size += 7
        assert isinstance(data.valid, bool)
        result |= ((int(data.valid)) & (2 ** (1) - 1)) << size
        size += 1
        isinstance(data.time_offset, (int, timedelta))
        time_offset_tmp2 = int(data.time_offset.total_seconds() // 1200 if isinstance(data.time_offset, timedelta) else data.time_offset // 1200)
        assert 0 <= time_offset_tmp2 <= 63
        result |= ((time_offset_tmp2) & (2 ** (6) - 1)) << size
        size += 6
        assert isinstance(data.journal, tuple) and len(data.journal) == 8
        isinstance(data.journal[0].offset, (int, timedelta))
        offset_tmp3 = int(data.journal[0].offset.total_seconds() // 60 if isinstance(data.journal[0].offset, timedelta) else data.journal[0].offset // 60)
        assert 0 <= offset_tmp3 <= 31
        result |= ((offset_tmp3) & (2 ** (5) - 1)) << size
        size += 5
        assert isinstance(data.journal[0].code, SmpmUlDeviceEnergy16bJournalType)
        result |= ((data.journal[0].code.value) & (2 ** (8) - 1)) << size
        size += 8
        isinstance(data.journal[1].offset, (int, timedelta))
        offset_tmp4 = int(data.journal[1].offset.total_seconds() // 60 if isinstance(data.journal[1].offset, timedelta) else data.journal[1].offset // 60)
        assert 0 <= offset_tmp4 <= 31
        result |= ((offset_tmp4) & (2 ** (5) - 1)) << size
        size += 5
        assert isinstance(data.journal[1].code, SmpmUlDeviceEnergy16bJournalType)
        result |= ((data.journal[1].code.value) & (2 ** (8) - 1)) << size
        size += 8
        isinstance(data.journal[2].offset, (int, timedelta))
        offset_tmp5 = int(data.journal[2].offset.total_seconds() // 60 if isinstance(data.journal[2].offset, timedelta) else data.journal[2].offset // 60)
        assert 0 <= offset_tmp5 <= 31
        result |= ((offset_tmp5) & (2 ** (5) - 1)) << size
        size += 5
        assert isinstance(data.journal[2].code, SmpmUlDeviceEnergy16bJournalType)
        result |= ((data.journal[2].code.value) & (2 ** (8) - 1)) << size
        size += 8
        isinstance(data.journal[3].offset, (int, timedelta))
        offset_tmp6 = int(data.journal[3].offset.total_seconds() // 60 if isinstance(data.journal[3].offset, timedelta) else data.journal[3].offset // 60)
        assert 0 <= offset_tmp6 <= 31
        result |= ((offset_tmp6) & (2 ** (5) - 1)) << size
        size += 5
        assert isinstance(data.journal[3].code, SmpmUlDeviceEnergy16bJournalType)
        result |= ((data.journal[3].code.value) & (2 ** (8) - 1)) << size
        size += 8
        isinstance(data.journal[4].offset, (int, timedelta))
        offset_tmp7 = int(data.journal[4].offset.total_seconds() // 60 if isinstance(data.journal[4].offset, timedelta) else data.journal[4].offset // 60)
        assert 0 <= offset_tmp7 <= 31
        result |= ((offset_tmp7) & (2 ** (5) - 1)) << size
        size += 5
        assert isinstance(data.journal[4].code, SmpmUlDeviceEnergy16bJournalType)
        result |= ((data.journal[4].code.value) & (2 ** (8) - 1)) << size
        size += 8
        isinstance(data.journal[5].offset, (int, timedelta))
        offset_tmp8 = int(data.journal[5].offset.total_seconds() // 60 if isinstance(data.journal[5].offset, timedelta) else data.journal[5].offset // 60)
        assert 0 <= offset_tmp8 <= 31
        result |= ((offset_tmp8) & (2 ** (5) - 1)) << size
        size += 5
        assert isinstance(data.journal[5].code, SmpmUlDeviceEnergy16bJournalType)
        result |= ((data.journal[5].code.value) & (2 ** (8) - 1)) << size
        size += 8
        isinstance(data.journal[6].offset, (int, timedelta))
        offset_tmp9 = int(data.journal[6].offset.total_seconds() // 60 if isinstance(data.journal[6].offset, timedelta) else data.journal[6].offset // 60)
        assert 0 <= offset_tmp9 <= 31
        result |= ((offset_tmp9) & (2 ** (5) - 1)) << size
        size += 5
        assert isinstance(data.journal[6].code, SmpmUlDeviceEnergy16bJournalType)
        result |= ((data.journal[6].code.value) & (2 ** (8) - 1)) << size
        size += 8
        isinstance(data.journal[7].offset, (int, timedelta))
        offset_tmp10 = int(data.journal[7].offset.total_seconds() // 60 if isinstance(data.journal[7].offset, timedelta) else data.journal[7].offset // 60)
        assert 0 <= offset_tmp10 <= 31
        result |= ((offset_tmp10) & (2 ** (5) - 1)) << size
        size += 5
        assert isinstance(data.journal[7].code, SmpmUlDeviceEnergy16bJournalType)
        result |= ((data.journal[7].code.value) & (2 ** (8) - 1)) << size
        size += 8
        return result.to_bytes(16, "little")

    @classmethod
    def parse(cls, buf: BufRef) -> 'SmpmUlDeviceEnergy16BJournalData':
        result__el_tmp11: Dict[str, Any] = dict()
        if 115 != buf.shift(7):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 0 != buf.shift(1):
            raise ValueError("packet_type_id: buffer doesn't match value")
        result__el_tmp11["days_ago"] = timedelta(seconds=buf.shift(7) * 86400)
        result__el_tmp11["valid"] = bool(buf.shift(1))
        result__el_tmp11["time_offset"] = timedelta(seconds=buf.shift(6) * 1200)
        journal_tmp12: List[EventData] = []
        journal__item_tmp13__el_tmp14: Dict[str, Any] = dict()
        journal__item_tmp13__el_tmp14["offset"] = timedelta(seconds=buf.shift(5) * 60)
        journal__item_tmp13__el_tmp14["code"] = SmpmUlDeviceEnergy16bJournalType(buf.shift(8))
        journal__item_tmp13 = EventData(**journal__item_tmp13__el_tmp14)
        journal_tmp12.append(journal__item_tmp13)
        journal__item_tmp13__el_tmp15: Dict[str, Any] = dict()
        journal__item_tmp13__el_tmp15["offset"] = timedelta(seconds=buf.shift(5) * 60)
        journal__item_tmp13__el_tmp15["code"] = SmpmUlDeviceEnergy16bJournalType(buf.shift(8))
        journal__item_tmp13 = EventData(**journal__item_tmp13__el_tmp15)
        journal_tmp12.append(journal__item_tmp13)
        journal__item_tmp13__el_tmp16: Dict[str, Any] = dict()
        journal__item_tmp13__el_tmp16["offset"] = timedelta(seconds=buf.shift(5) * 60)
        journal__item_tmp13__el_tmp16["code"] = SmpmUlDeviceEnergy16bJournalType(buf.shift(8))
        journal__item_tmp13 = EventData(**journal__item_tmp13__el_tmp16)
        journal_tmp12.append(journal__item_tmp13)
        journal__item_tmp13__el_tmp17: Dict[str, Any] = dict()
        journal__item_tmp13__el_tmp17["offset"] = timedelta(seconds=buf.shift(5) * 60)
        journal__item_tmp13__el_tmp17["code"] = SmpmUlDeviceEnergy16bJournalType(buf.shift(8))
        journal__item_tmp13 = EventData(**journal__item_tmp13__el_tmp17)
        journal_tmp12.append(journal__item_tmp13)
        journal__item_tmp13__el_tmp18: Dict[str, Any] = dict()
        journal__item_tmp13__el_tmp18["offset"] = timedelta(seconds=buf.shift(5) * 60)
        journal__item_tmp13__el_tmp18["code"] = SmpmUlDeviceEnergy16bJournalType(buf.shift(8))
        journal__item_tmp13 = EventData(**journal__item_tmp13__el_tmp18)
        journal_tmp12.append(journal__item_tmp13)
        journal__item_tmp13__el_tmp19: Dict[str, Any] = dict()
        journal__item_tmp13__el_tmp19["offset"] = timedelta(seconds=buf.shift(5) * 60)
        journal__item_tmp13__el_tmp19["code"] = SmpmUlDeviceEnergy16bJournalType(buf.shift(8))
        journal__item_tmp13 = EventData(**journal__item_tmp13__el_tmp19)
        journal_tmp12.append(journal__item_tmp13)
        journal__item_tmp13__el_tmp20: Dict[str, Any] = dict()
        journal__item_tmp13__el_tmp20["offset"] = timedelta(seconds=buf.shift(5) * 60)
        journal__item_tmp13__el_tmp20["code"] = SmpmUlDeviceEnergy16bJournalType(buf.shift(8))
        journal__item_tmp13 = EventData(**journal__item_tmp13__el_tmp20)
        journal_tmp12.append(journal__item_tmp13)
        journal__item_tmp13__el_tmp21: Dict[str, Any] = dict()
        journal__item_tmp13__el_tmp21["offset"] = timedelta(seconds=buf.shift(5) * 60)
        journal__item_tmp13__el_tmp21["code"] = SmpmUlDeviceEnergy16bJournalType(buf.shift(8))
        journal__item_tmp13 = EventData(**journal__item_tmp13__el_tmp21)
        journal_tmp12.append(journal__item_tmp13)
        result__el_tmp11["journal"] = tuple(journal_tmp12)
        result = SmpmUlDeviceEnergy16BJournalData(**result__el_tmp11)
        buf.shift(2)
        return result

    def to_integration_data(self, received_at: datetime, device_tz: tzinfo, **kwargs: Any) -> List[IntegrationV0MessageData]:
        date = days_ago_calculation(received_at, device_tz, received_at.time(), self.days_ago)
        dt = datetime.combine(date.date(), (date + self.time_offset).time()) if self.days_ago.total_seconds() else datetime.combine(received_at.date(), time(hour=0, minute=0, second=0))   # noqa: E501
        integration_messages = []
        for event in self.journal:
            if event.code is SmpmUlDeviceEnergy16bJournalType.NONE:
                continue
            integration_messages.append(IntegrationV0MessageData(
                is_valid=self.valid,
                dt=dt + event.offset,
                events=[EVENTS_MAP[event.code]],
            ))
        return integration_messages



import importlib
import os.path
from glob import glob

CWD = os.path.abspath(os.path.join(os.path.dirname(__file__), '../', '../', '../', '../'))
PARENT_DIR = os.path.abspath(os.path.join(os.path.dirname(__file__), '../'))
MDL_PREF = os.path.relpath(PARENT_DIR, CWD).replace('/', '.').strip('.')


def test_some() -> None:
    for fn in glob(os.path.join(PARENT_DIR, '*.py')):
        if not os.path.isfile(fn) or '__' in fn:
            continue
        try:
            mdl = importlib.import_module(f'{MDL_PREF}.{os.path.basename(fn)[:-3]}')
        except Exception:  # noqa: B902
            continue
        for el in dir(mdl):
            if not el.startswith('test_'):
                continue
            getattr(mdl, el)()  # run test functions from each pack
    assert True



from data_gateway_sdk.protocols.water5_v_nero.water5_device_daily import Water5DeviceDailyData
from data_gateway_sdk.utils.buf_ref import BufRef


def test_water5_device_daily() -> None:
    case = Water5DeviceDailyData(value=8.3)  # noqa: E501
    assert case == Water5DeviceDailyData.parse(BufRef(case.serialize()))
    case = Water5DeviceDailyData(value=0.0)  # noqa: E501
    assert case == Water5DeviceDailyData.parse(BufRef(case.serialize()))
    case = Water5DeviceDailyData(value=8.3)  # noqa: E501
    assert case == Water5DeviceDailyData.parse(BufRef(case.serialize()))
    case = Water5DeviceDailyData(value=32.767)  # noqa: E501
    assert case == Water5DeviceDailyData.parse(BufRef(case.serialize()))



from decimal import Decimal
from datetime import timedelta, tzinfo, datetime
from typing import Dict, Any, List

from data_aggregator_sdk.constants.enums import SensorType
from data_aggregator_sdk.integration_message import IntegrationV0MessageData, IntegrationV0MessageRelay, IntegrationV0MessageClock, IntegrationV0MessageSensor

from data_gateway_sdk.utils.buf_ref import BufRef
from data_gateway_sdk.utils.packet import Packet


# PACKET (128 bits)   smpm_ul_device_energy_16b_status_info
#
# RESULT int:        114624888726607593574237117
# RESULT bin:  MSB   00000000 00000000 00000000 00000000 00000000 01011110 11010000 11000011 10101011 00111110 00000001 00001110 10101100 11111000 00000011 10111101   LSB
# RESULT hex:  LE    bd 03 f8 ac 0e 01 3e ab c3 d0 5e 00 00 00 00 00
#
# name                  type       size  value(int)                                                                                                                        data(bits)
# --------------------  ---------  ----  ----------  --------------------------------------------------------------------------------------------------------------------------------
# packet_type_id.0.VAL  u7            7          61                                                                                                                           0111101
# packet_type_id.0.DFF  bool          1           1                                                                                                                          1
# packet_type_id.1.VAL  u2            2           3                                                                                                                        11
# packet_type_id.1.DFF  bool          1           0                                                                                                                       0
# datetime              timedelta    30   567648000                                                                                         100001110101011001111100000000
# uptime_s              timedelta    30   567648000                                                           100001110101011001111100000000
# battery_volts         uf7p1         7          33                                                    0100001
# temperature           u8            8         123                                            01111011
# relay_switch          u2            2           1                                          01
# relay_is_active       bool          1           0                                         0
# RESERVED              u39          39           0  000000000000000000000000000000000000000


class SmpmUlDeviceEnergy16BStatusInfoData(Packet):
    date_time: timedelta
    uptime_s: timedelta
    battery_volts: float
    temperature: int
    relay_switch: int
    relay_is_active: bool

    def serialize(self) -> bytes:
        data = self
        result = 0
        size = 0
        result |= ((61) & (2 ** (7) - 1)) << size
        size += 7
        result |= ((1) & (2 ** (1) - 1)) << size
        size += 1
        result |= ((3) & (2 ** (2) - 1)) << size
        size += 2
        result |= ((0) & (2 ** (1) - 1)) << size
        size += 1
        isinstance(data.date_time, (int, timedelta))
        datetime_tmp1 = int(data.date_time.total_seconds() // 1 if isinstance(data.date_time, timedelta) else data.date_time // 1)
        assert 0 <= datetime_tmp1 <= 1073741823
        result |= ((datetime_tmp1) & (2 ** (30) - 1)) << size
        size += 30
        isinstance(data.uptime_s, (int, timedelta))
        value_int_tmp2 = int(data.uptime_s.total_seconds() // 1 if isinstance(data.uptime_s, timedelta) else data.uptime_s // 1) & 1073741823
        result |= ((value_int_tmp2) & (2 ** (30) - 1)) << size
        size += 30
        assert isinstance(data.battery_volts, (int, float))
        assert 0.0 <= data.battery_volts <= 12.7
        result |= ((int(round(float(data.battery_volts) * 10.0, 0))) & (2 ** (7) - 1)) << size
        size += 7
        assert isinstance(data.temperature, int)
        assert -100 <= data.temperature <= 155
        result |= (((data.temperature + 100)) & (2 ** (8) - 1)) << size
        size += 8
        assert isinstance(data.relay_switch, int)
        assert 0 <= data.relay_switch <= 3
        result |= ((data.relay_switch) & (2 ** (2) - 1)) << size
        size += 2
        assert isinstance(data.relay_is_active, bool)
        result |= ((int(data.relay_is_active)) & (2 ** (1) - 1)) << size
        size += 1
        return result.to_bytes(16, "little")

    @classmethod
    def parse(cls, buf: BufRef) -> 'SmpmUlDeviceEnergy16BStatusInfoData':
        result__el_tmp3: Dict[str, Any] = dict()
        if 61 != buf.shift(7):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 1 != buf.shift(1):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 3 != buf.shift(2):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 0 != buf.shift(1):
            raise ValueError("packet_type_id: buffer doesn't match value")
        result__el_tmp3["date_time"] = timedelta(seconds=buf.shift(30) * 1)
        result__el_tmp3["uptime_s"] = timedelta(seconds=buf.shift(30) * 1)
        result__el_tmp3["battery_volts"] = round(buf.shift(7) / 10.0, 1)
        result__el_tmp3["temperature"] = buf.shift(8) + -100
        result__el_tmp3["relay_switch"] = buf.shift(2) + 0
        result__el_tmp3["relay_is_active"] = bool(buf.shift(1))
        result = SmpmUlDeviceEnergy16BStatusInfoData(**result__el_tmp3)
        buf.shift(39)
        return result

    def to_integration_data(self, received_at: datetime, device_tz: tzinfo, **kwargs: Any) -> List[IntegrationV0MessageData]:
        return [
            IntegrationV0MessageData(
                dt=received_at,
                sensors=[
                    IntegrationV0MessageSensor(
                        value=Decimal(str(self.battery_volts)),
                        sensor_type=SensorType.BATTERY,
                    ),
                    IntegrationV0MessageSensor(
                        value=Decimal(str(self.temperature)),
                        sensor_type=SensorType.TEMPERATURE,
                    ),
                    IntegrationV0MessageSensor(
                        value=Decimal(str(self.uptime_s.total_seconds())),
                        sensor_type=SensorType.UPTIME,
                    ),
                ],
                relay=[
                    IntegrationV0MessageRelay(value=self.relay_is_active),
                ],
                clock=[
                    IntegrationV0MessageClock(value=datetime(year=2020, month=1, day=1) + self.date_time),
                ],
            ),
        ]



from datetime import timedelta

from data_gateway_sdk.protocols.smpm.smpm_ul_device_energy_16b_profile_8h_energy import SmpmUlDeviceEnergy16BProfile8HEnergyData, SmpmUlDeviceEnergy16bProfile8hEnergyIds
from data_gateway_sdk.utils.buf_ref import BufRef


def test_smpm_ul_device_energy_16b_profile_8h_energy() -> None:
    case_serialized = bytes.fromhex("b0035500200008800140000a80013800")
    assert SmpmUlDeviceEnergy16BProfile8HEnergyData(packet_type_id_enum=SmpmUlDeviceEnergy16bProfile8hEnergyIds.UL_DATA_16B__PROFILE_ENERGY_ACTIVE_CONSUMED_H8_1, days_ago=timedelta(seconds=0.0), valid=(True, False, True, False, True, False, True, False), profile=(0, 1, 2, 3, 4, 5, 6, 7)) == SmpmUlDeviceEnergy16BProfile8HEnergyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BProfile8HEnergyData.serialize(SmpmUlDeviceEnergy16BProfile8HEnergyData(packet_type_id_enum=SmpmUlDeviceEnergy16bProfile8hEnergyIds.UL_DATA_16B__PROFILE_ENERGY_ACTIVE_CONSUMED_H8_1, days_ago=timedelta(seconds=0.0), valid=(True, False, True, False, True, False, True, False), profile=(0, 1, 2, 3, 4, 5, 6, 7)))  # noqa: E501
    case_serialized = bytes.fromhex("b0035500200008800140000a80013800")
    assert SmpmUlDeviceEnergy16BProfile8HEnergyData(packet_type_id_enum=SmpmUlDeviceEnergy16bProfile8hEnergyIds.UL_DATA_16B__PROFILE_ENERGY_ACTIVE_CONSUMED_H8_1, days_ago=timedelta(seconds=0.0), valid=(True, False, True, False, True, False, True, False), profile=(0, 1, 2, 3, 4, 5, 6, 7)) == SmpmUlDeviceEnergy16BProfile8HEnergyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BProfile8HEnergyData.serialize(SmpmUlDeviceEnergy16BProfile8HEnergyData(days_ago=timedelta(seconds=0.0), packet_type_id_enum=SmpmUlDeviceEnergy16bProfile8hEnergyIds.UL_DATA_16B__PROFILE_ENERGY_ACTIVE_CONSUMED_H8_1, profile=(0, 1, 2, 3, 4, 5, 6, 7), valid=(True, False, True, False, True, False, True, False)))  # noqa: E501
    case_serialized = bytes.fromhex("bb035500200008800140000a80013800")
    assert SmpmUlDeviceEnergy16BProfile8HEnergyData(packet_type_id_enum=SmpmUlDeviceEnergy16bProfile8hEnergyIds.UL_DATA_16B__PROFILE_ENERGY_REACTIVE_GENERATED_H8_3, days_ago=timedelta(seconds=0.0), valid=(True, False, True, False, True, False, True, False), profile=(0, 1, 2, 3, 4, 5, 6, 7)) == SmpmUlDeviceEnergy16BProfile8HEnergyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BProfile8HEnergyData.serialize(SmpmUlDeviceEnergy16BProfile8HEnergyData(days_ago=timedelta(seconds=0.0), packet_type_id_enum=SmpmUlDeviceEnergy16bProfile8hEnergyIds.UL_DATA_16B__PROFILE_ENERGY_REACTIVE_GENERATED_H8_3, profile=(0, 1, 2, 3, 4, 5, 6, 7), valid=(True, False, True, False, True, False, True, False)))  # noqa: E501
    case_serialized = bytes.fromhex("b0fb5500200008800140000a80013800")
    assert SmpmUlDeviceEnergy16BProfile8HEnergyData(packet_type_id_enum=SmpmUlDeviceEnergy16bProfile8hEnergyIds.UL_DATA_16B__PROFILE_ENERGY_ACTIVE_CONSUMED_H8_1, days_ago=timedelta(seconds=2678400.0), valid=(True, False, True, False, True, False, True, False), profile=(0, 1, 2, 3, 4, 5, 6, 7)) == SmpmUlDeviceEnergy16BProfile8HEnergyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BProfile8HEnergyData.serialize(SmpmUlDeviceEnergy16BProfile8HEnergyData(days_ago=timedelta(seconds=2678400.0), packet_type_id_enum=SmpmUlDeviceEnergy16bProfile8hEnergyIds.UL_DATA_16B__PROFILE_ENERGY_ACTIVE_CONSUMED_H8_1, profile=(0, 1, 2, 3, 4, 5, 6, 7), valid=(True, False, True, False, True, False, True, False)))  # noqa: E501
    case_serialized = bytes.fromhex("bbfb5500200008800140000a80013800")
    assert SmpmUlDeviceEnergy16BProfile8HEnergyData(packet_type_id_enum=SmpmUlDeviceEnergy16bProfile8hEnergyIds.UL_DATA_16B__PROFILE_ENERGY_REACTIVE_GENERATED_H8_3, days_ago=timedelta(seconds=2678400.0), valid=(True, False, True, False, True, False, True, False), profile=(0, 1, 2, 3, 4, 5, 6, 7)) == SmpmUlDeviceEnergy16BProfile8HEnergyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BProfile8HEnergyData.serialize(SmpmUlDeviceEnergy16BProfile8HEnergyData(days_ago=timedelta(seconds=2678400.0), packet_type_id_enum=SmpmUlDeviceEnergy16bProfile8hEnergyIds.UL_DATA_16B__PROFILE_ENERGY_REACTIVE_GENERATED_H8_3, profile=(0, 1, 2, 3, 4, 5, 6, 7), valid=(True, False, True, False, True, False, True, False)))  # noqa: E501



from datetime import timedelta

from data_gateway_sdk.protocols.smpm.smpm_ul_device_energy_16b_profile_8h_ar import SmpmUlDeviceEnergy16BProfile8HArData, SmpmUlDeviceEnergy16bProfile8hARIds
from data_gateway_sdk.utils.buf_ref import BufRef


def test_smpm_ul_device_energy_16b_profile_8h_ar() -> None:
    case_serialized = bytes.fromhex("810c0000012000034000056000072003")
    assert SmpmUlDeviceEnergy16BProfile8HArData(packet_type_id=SmpmUlDeviceEnergy16bProfile8hARIds.UL_DATA_16B__PROFILE_H8_1_AP, days_ago=timedelta(seconds=0.0), profile=(0, 1, 2, 3, 4, 5, 6, 7), point_factor=5.0) == SmpmUlDeviceEnergy16BProfile8HArData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BProfile8HArData.serialize(SmpmUlDeviceEnergy16BProfile8HArData(packet_type_id=SmpmUlDeviceEnergy16bProfile8hARIds.UL_DATA_16B__PROFILE_H8_1_AP, days_ago=timedelta(seconds=0.0), profile=(0, 1, 2, 3, 4, 5, 6, 7), point_factor=5.0))  # noqa: E501
    case_serialized = bytes.fromhex("b00c0000012000034000056000070000")
    assert SmpmUlDeviceEnergy16BProfile8HArData(packet_type_id=SmpmUlDeviceEnergy16bProfile8hARIds.UL_DATA_16B__PROFILE_H8_3_RN__PHASE_C, days_ago=timedelta(seconds=0.0), profile=(0, 1, 2, 3, 4, 5, 6, 7), point_factor=0.0) == SmpmUlDeviceEnergy16BProfile8HArData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BProfile8HArData.serialize(SmpmUlDeviceEnergy16BProfile8HArData(days_ago=timedelta(seconds=0.0), packet_type_id=SmpmUlDeviceEnergy16bProfile8hARIds.UL_DATA_16B__PROFILE_H8_3_RN__PHASE_C, point_factor=0.0, profile=(0, 1, 2, 3, 4, 5, 6, 7)))  # noqa: E501
    case_serialized = bytes.fromhex("b00c000001200003400005600007f0ff")
    assert SmpmUlDeviceEnergy16BProfile8HArData(packet_type_id=SmpmUlDeviceEnergy16bProfile8hARIds.UL_DATA_16B__PROFILE_H8_3_RN__PHASE_C, days_ago=timedelta(seconds=0.0), profile=(0, 1, 2, 3, 4, 5, 6, 7), point_factor=409.5) == SmpmUlDeviceEnergy16BProfile8HArData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BProfile8HArData.serialize(SmpmUlDeviceEnergy16BProfile8HArData(days_ago=timedelta(seconds=0.0), packet_type_id=SmpmUlDeviceEnergy16bProfile8hARIds.UL_DATA_16B__PROFILE_H8_3_RN__PHASE_C, point_factor=409.5, profile=(0, 1, 2, 3, 4, 5, 6, 7)))  # noqa: E501
    case_serialized = bytes.fromhex("810c0000012000034000056000070000")
    assert SmpmUlDeviceEnergy16BProfile8HArData(packet_type_id=SmpmUlDeviceEnergy16bProfile8hARIds.UL_DATA_16B__PROFILE_H8_1_AP, days_ago=timedelta(seconds=0.0), profile=(0, 1, 2, 3, 4, 5, 6, 7), point_factor=0.0) == SmpmUlDeviceEnergy16BProfile8HArData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BProfile8HArData.serialize(SmpmUlDeviceEnergy16BProfile8HArData(days_ago=timedelta(seconds=0.0), packet_type_id=SmpmUlDeviceEnergy16bProfile8hARIds.UL_DATA_16B__PROFILE_H8_1_AP, point_factor=0.0, profile=(0, 1, 2, 3, 4, 5, 6, 7)))  # noqa: E501
    case_serialized = bytes.fromhex("810c000001200003400005600007f0ff")
    assert SmpmUlDeviceEnergy16BProfile8HArData(packet_type_id=SmpmUlDeviceEnergy16bProfile8hARIds.UL_DATA_16B__PROFILE_H8_1_AP, days_ago=timedelta(seconds=0.0), profile=(0, 1, 2, 3, 4, 5, 6, 7), point_factor=409.5) == SmpmUlDeviceEnergy16BProfile8HArData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BProfile8HArData.serialize(SmpmUlDeviceEnergy16BProfile8HArData(days_ago=timedelta(seconds=0.0), packet_type_id=SmpmUlDeviceEnergy16bProfile8hARIds.UL_DATA_16B__PROFILE_H8_1_AP, point_factor=409.5, profile=(0, 1, 2, 3, 4, 5, 6, 7)))  # noqa: E501
    case_serialized = bytes.fromhex("b0cc0f00012000034000056000070000")
    assert SmpmUlDeviceEnergy16BProfile8HArData(packet_type_id=SmpmUlDeviceEnergy16bProfile8hARIds.UL_DATA_16B__PROFILE_H8_3_RN__PHASE_C, days_ago=timedelta(seconds=5443200.0), profile=(0, 1, 2, 3, 4, 5, 6, 7), point_factor=0.0) == SmpmUlDeviceEnergy16BProfile8HArData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BProfile8HArData.serialize(SmpmUlDeviceEnergy16BProfile8HArData(days_ago=timedelta(seconds=5443200.0), packet_type_id=SmpmUlDeviceEnergy16bProfile8hARIds.UL_DATA_16B__PROFILE_H8_3_RN__PHASE_C, point_factor=0.0, profile=(0, 1, 2, 3, 4, 5, 6, 7)))  # noqa: E501
    case_serialized = bytes.fromhex("b0cc0f0001200003400005600007f0ff")
    assert SmpmUlDeviceEnergy16BProfile8HArData(packet_type_id=SmpmUlDeviceEnergy16bProfile8hARIds.UL_DATA_16B__PROFILE_H8_3_RN__PHASE_C, days_ago=timedelta(seconds=5443200.0), profile=(0, 1, 2, 3, 4, 5, 6, 7), point_factor=409.5) == SmpmUlDeviceEnergy16BProfile8HArData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BProfile8HArData.serialize(SmpmUlDeviceEnergy16BProfile8HArData(days_ago=timedelta(seconds=5443200.0), packet_type_id=SmpmUlDeviceEnergy16bProfile8hARIds.UL_DATA_16B__PROFILE_H8_3_RN__PHASE_C, point_factor=409.5, profile=(0, 1, 2, 3, 4, 5, 6, 7)))  # noqa: E501
    case_serialized = bytes.fromhex("81cc0f00012000034000056000070000")
    assert SmpmUlDeviceEnergy16BProfile8HArData(packet_type_id=SmpmUlDeviceEnergy16bProfile8hARIds.UL_DATA_16B__PROFILE_H8_1_AP, days_ago=timedelta(seconds=5443200.0), profile=(0, 1, 2, 3, 4, 5, 6, 7), point_factor=0.0) == SmpmUlDeviceEnergy16BProfile8HArData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BProfile8HArData.serialize(SmpmUlDeviceEnergy16BProfile8HArData(days_ago=timedelta(seconds=5443200.0), packet_type_id=SmpmUlDeviceEnergy16bProfile8hARIds.UL_DATA_16B__PROFILE_H8_1_AP, point_factor=0.0, profile=(0, 1, 2, 3, 4, 5, 6, 7)))  # noqa: E501
    case_serialized = bytes.fromhex("81cc0f0001200003400005600007f0ff")
    assert SmpmUlDeviceEnergy16BProfile8HArData(packet_type_id=SmpmUlDeviceEnergy16bProfile8hARIds.UL_DATA_16B__PROFILE_H8_1_AP, days_ago=timedelta(seconds=5443200.0), profile=(0, 1, 2, 3, 4, 5, 6, 7), point_factor=409.5) == SmpmUlDeviceEnergy16BProfile8HArData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BProfile8HArData.serialize(SmpmUlDeviceEnergy16BProfile8HArData(days_ago=timedelta(seconds=5443200.0), packet_type_id=SmpmUlDeviceEnergy16bProfile8hARIds.UL_DATA_16B__PROFILE_H8_1_AP, point_factor=409.5, profile=(0, 1, 2, 3, 4, 5, 6, 7)))  # noqa: E501



from decimal import Decimal
from datetime import timedelta, datetime, time, tzinfo
from typing import List, Any, Dict

from data_aggregator_sdk.constants.enums import JournalDataType, SensorType
from data_aggregator_sdk.integration_message import IntegrationV0MessageEvent, IntegrationV0MessageData, \
    IntegrationV0MessageConsumption, CounterType, ResourceType, IntegrationV0MessageGeneration, IntegrationV0MessageClock, IntegrationV0MessageSensor, \
    IntegrationV0MessageConnectionQuality
from pytz import timezone

from data_gateway_sdk.utils.buf_ref import BufRef
from data_gateway_sdk.utils.days_ago_calculation import days_ago_calculation
from data_gateway_sdk.utils.packet import Packet
from data_gateway_sdk.utils.timestamp_calculation import timestamp_calculation


# PACKET (256 bits)   smpm_ul_device_gaz_flow_32b_daily
#
# RESULT int:        1759786798423479316420289521988439236689127852665267125398871573399001369732
# RESULT bin:  MSB   00000011 11100100 00000001 00001000 00101010 10000110 00011101 10110110 00000000 00001010 00000000 00001010 00000001 10111000 00001101 01111100 00001000 00001010 00100101 00100011 00000110 10110001 11101010 11100100 10010000 10111010 01111111 11111111 11111111 11000000 00001100 10000100   LSB
# RESULT hex:  LE    84 0c c0 ff ff 7f ba 90 e4 ea b1 06 23 25 0a 08 7c 0d b8 01 0a 00 0a 00 b6 1d 86 2a 08 01 e4 03
#
# name                            type       size  value(int)                                                                                                                                                                                                                                                        data(bits)
# ------------------------------  ---------  ----  ----------  ----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
# packet_type_id.0.VAL            u7            7           4                                                                                                                                                                                                                                                           0000100
# packet_type_id.0.DFF            bool          1           1                                                                                                                                                                                                                                                          1
# packet_type_id.1.VAL            u2            2           0                                                                                                                                                                                                                                                        00
# packet_type_id.1.DFF            bool          1           1                                                                                                                                                                                                                                                       1
# packet_type_id.2.VAL            u2            2           1                                                                                                                                                                                                                                                     01
# packet_type_id.2.DFF            bool          1           0                                                                                                                                                                                                                                                    0
# days_ago                        timedelta     5           0                                                                                                                                                                                                                                               00000
# sync_time_days_ago              timedelta     3           0                                                                                                                                                                                                                                            000
# timestamp_s                     timedelta    26    33554431                                                                                                                                                                                                                  01111111111111111111111111
# temperature                     u7            7          58                                                                                                                                                                                                           0111010
# battery_volts                   uf6p1         6          33                                                                                                                                                                                                     100001
# event_reset                     bool          1           0                                                                                                                                                                                                    0
# event_low_battery_level         bool          1           0                                                                                                                                                                                                   0
# event_temperature_limits        bool          1           1                                                                                                                                                                                                  1
# direct_flow_volume              uf32p3       32   112323300                                                                                                                                                                  00000110101100011110101011100100
# direct_flow_volume_day_ago      uf7p1         7          35                                                                                                                                                           0100011
# reverse_flow_volume             uf12p2       12        1098                                                                                                                                               010001001010
# event_battery_warn              bool          1           1                                                                                                                                              1
# event_system_error              bool          1           0                                                                                                                                             0
# event_flow_reverse              bool          1           0                                                                                                                                            0
# event_flow_speed_is_over_limit  bool          1           0                                                                                                                                           0
# event_sensor_error              bool          1           0                                                                                                                                          0
# event_sensor_error_temperature  bool          1           0                                                                                                                                         0
# event_case_was_opened           bool          1           0                                                                                                                                        0
# event_continuous_consumption    bool          1           0                                                                                                                                       0
# event_no_resource               bool          1           1                                                                                                                                      1
# RESERVED                        u4            4           0                                                                                                                                  0000
# earfcn                          u16          16        3452                                                                                                                  0000110101111100
# pci                             u16          16         440                                                                                                  0000000110111000
# cell_id                         u16          16          10                                                                                  0000000000001010
# tac                             u16          16          10                                                                  0000000000001010
# rsrp                            u8            8         182                                                          10110110
# rsrq                            u8            8          29                                                  00011101
# rssi                            u8            8         134                                          10000110
# snr                             u8            8          42                                  00101010
# band                            u8            8           8                          00001000
# ecl                             u8            8           1                  00000001
# tx_power                        u8            8         228          11100100
# operation_mode                  u8            8           3  00000011


class SmpmUlDeviceGazFlow32BDailyData(Packet):
    days_ago: timedelta
    sync_time_days_ago: timedelta
    timestamp_s: timedelta
    temperature: int
    battery_volts: float
    event_reset: bool
    event_low_battery_level: bool
    event_temperature_limits: bool
    direct_flow_volume: float
    direct_flow_volume_day_ago: float
    reverse_flow_volume: float
    event_battery_warn: bool
    event_system_error: bool
    event_flow_reverse: bool
    event_flow_speed_is_over_limit: bool
    event_sensor_error: bool
    event_sensor_error_temperature: bool
    event_case_was_opened: bool
    event_continuous_consumption: bool
    event_no_resource: bool
    earfcn: int
    pci: int
    cell_id: int
    tac: int
    rsrp: int
    rsrq: int
    rssi: int
    snr: int
    band: int
    ecl: int
    tx_power: int
    operation_mode: int

    def serialize(self) -> bytes:
        data = self
        result = 0
        size = 0
        result |= ((4) & (2 ** (7) - 1)) << size
        size += 7
        result |= ((1) & (2 ** (1) - 1)) << size
        size += 1
        result |= ((0) & (2 ** (2) - 1)) << size
        size += 2
        result |= ((1) & (2 ** (1) - 1)) << size
        size += 1
        result |= ((1) & (2 ** (2) - 1)) << size
        size += 2
        result |= ((0) & (2 ** (1) - 1)) << size
        size += 1
        isinstance(data.days_ago, (int, timedelta))
        days_ago_tmp1 = int(data.days_ago.total_seconds() // 86400 if isinstance(data.days_ago, timedelta) else data.days_ago // 86400)
        assert 0 <= days_ago_tmp1 <= 31
        result |= ((days_ago_tmp1) & (2 ** (5) - 1)) << size
        size += 5
        isinstance(data.sync_time_days_ago, (int, timedelta))
        result |= ((max(min(7, int(data.sync_time_days_ago.total_seconds() // 86400 if isinstance(data.sync_time_days_ago, timedelta) else data.sync_time_days_ago // 86400)), 0)) & (2 ** (3) - 1)) << size
        size += 3
        isinstance(data.timestamp_s, (int, timedelta))
        value_int_tmp2 = int(data.timestamp_s.total_seconds() // 1 if isinstance(data.timestamp_s, timedelta) else data.timestamp_s // 1) & 67108863
        result |= ((value_int_tmp2) & (2 ** (26) - 1)) << size
        size += 26
        assert isinstance(data.temperature, int)
        assert -35 <= data.temperature <= 92
        result |= (((data.temperature + 35)) & (2 ** (7) - 1)) << size
        size += 7
        assert isinstance(data.battery_volts, (int, float))
        assert 0.0 <= data.battery_volts <= 6.3
        result |= ((int(round(float(data.battery_volts) * 10.0, 0))) & (2 ** (6) - 1)) << size
        size += 6
        assert isinstance(data.event_reset, bool)
        result |= ((int(data.event_reset)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_low_battery_level, bool)
        result |= ((int(data.event_low_battery_level)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_temperature_limits, bool)
        result |= ((int(data.event_temperature_limits)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.direct_flow_volume, (int, float))
        result |= ((int(round(float(data.direct_flow_volume) * 1000.0, 0)) & 4294967295) & (2 ** (32) - 1)) << size
        size += 32
        assert isinstance(data.direct_flow_volume_day_ago, (int, float))
        result |= ((int(round(max(min(12.7, float(data.direct_flow_volume_day_ago)), 0.0) * 10.0, 0))) & (2 ** (7) - 1)) << size
        size += 7
        assert isinstance(data.reverse_flow_volume, (int, float))
        result |= ((int(round(float(data.reverse_flow_volume) * 100.0, 0)) & 4095) & (2 ** (12) - 1)) << size
        size += 12
        assert isinstance(data.event_battery_warn, bool)
        result |= ((int(data.event_battery_warn)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_system_error, bool)
        result |= ((int(data.event_system_error)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_flow_reverse, bool)
        result |= ((int(data.event_flow_reverse)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_flow_speed_is_over_limit, bool)
        result |= ((int(data.event_flow_speed_is_over_limit)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_sensor_error, bool)
        result |= ((int(data.event_sensor_error)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_sensor_error_temperature, bool)
        result |= ((int(data.event_sensor_error_temperature)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_case_was_opened, bool)
        result |= ((int(data.event_case_was_opened)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_continuous_consumption, bool)
        result |= ((int(data.event_continuous_consumption)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_no_resource, bool)
        result |= ((int(data.event_no_resource)) & (2 ** (1) - 1)) << size
        size += 1
        result |= ((0) & (2 ** (4) - 1)) << size
        size += 4
        assert isinstance(data.earfcn, int)
        assert 0 <= data.earfcn <= 65535
        result |= ((data.earfcn) & (2 ** (16) - 1)) << size
        size += 16
        assert isinstance(data.pci, int)
        assert 0 <= data.pci <= 65535
        result |= ((data.pci) & (2 ** (16) - 1)) << size
        size += 16
        assert isinstance(data.cell_id, int)
        assert 0 <= data.cell_id <= 65535
        result |= ((data.cell_id) & (2 ** (16) - 1)) << size
        size += 16
        assert isinstance(data.tac, int)
        assert 0 <= data.tac <= 65535
        result |= ((data.tac) & (2 ** (16) - 1)) << size
        size += 16
        assert isinstance(data.rsrp, int)
        assert -140 <= data.rsrp <= 115
        result |= (((data.rsrp + 140)) & (2 ** (8) - 1)) << size
        size += 8
        assert isinstance(data.rsrq, int)
        assert -20 <= data.rsrq <= 235
        result |= (((data.rsrq + 20)) & (2 ** (8) - 1)) << size
        size += 8
        assert isinstance(data.rssi, int)
        assert -110 <= data.rssi <= 145
        result |= (((data.rssi + 110)) & (2 ** (8) - 1)) << size
        size += 8
        assert isinstance(data.snr, int)
        assert -20 <= data.snr <= 235
        result |= (((data.snr + 20)) & (2 ** (8) - 1)) << size
        size += 8
        assert isinstance(data.band, int)
        assert 0 <= data.band <= 255
        result |= ((data.band) & (2 ** (8) - 1)) << size
        size += 8
        assert isinstance(data.ecl, int)
        assert 0 <= data.ecl <= 255
        result |= ((data.ecl) & (2 ** (8) - 1)) << size
        size += 8
        assert isinstance(data.tx_power, int)
        assert -128 <= data.tx_power <= 127
        result |= (((data.tx_power + 128)) & (2 ** (8) - 1)) << size
        size += 8
        assert isinstance(data.operation_mode, int)
        assert 0 <= data.operation_mode <= 255
        result |= ((data.operation_mode) & (2 ** (8) - 1)) << size
        size += 8
        return result.to_bytes(32, "little")

    @classmethod
    def parse(cls, buf: BufRef) -> 'SmpmUlDeviceGazFlow32BDailyData':
        result__el_tmp3: Dict[str, Any] = dict()
        if 4 != buf.shift(7):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 1 != buf.shift(1):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 0 != buf.shift(2):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 1 != buf.shift(1):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 1 != buf.shift(2):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 0 != buf.shift(1):
            raise ValueError("packet_type_id: buffer doesn't match value")
        result__el_tmp3["days_ago"] = timedelta(seconds=buf.shift(5) * 86400)
        result__el_tmp3["sync_time_days_ago"] = timedelta(seconds=buf.shift(3) * 86400)
        result__el_tmp3["timestamp_s"] = timedelta(seconds=buf.shift(26) * 1)
        result__el_tmp3["temperature"] = buf.shift(7) + -35
        result__el_tmp3["battery_volts"] = round(buf.shift(6) / 10.0, 1)
        result__el_tmp3["event_reset"] = bool(buf.shift(1))
        result__el_tmp3["event_low_battery_level"] = bool(buf.shift(1))
        result__el_tmp3["event_temperature_limits"] = bool(buf.shift(1))
        result__el_tmp3["direct_flow_volume"] = round(buf.shift(32) / 1000.0, 3)
        result__el_tmp3["direct_flow_volume_day_ago"] = round(buf.shift(7) / 10.0, 1)
        result__el_tmp3["reverse_flow_volume"] = round(buf.shift(12) / 100.0, 2)
        result__el_tmp3["event_battery_warn"] = bool(buf.shift(1))
        result__el_tmp3["event_system_error"] = bool(buf.shift(1))
        result__el_tmp3["event_flow_reverse"] = bool(buf.shift(1))
        result__el_tmp3["event_flow_speed_is_over_limit"] = bool(buf.shift(1))
        result__el_tmp3["event_sensor_error"] = bool(buf.shift(1))
        result__el_tmp3["event_sensor_error_temperature"] = bool(buf.shift(1))
        result__el_tmp3["event_case_was_opened"] = bool(buf.shift(1))
        result__el_tmp3["event_continuous_consumption"] = bool(buf.shift(1))
        result__el_tmp3["event_no_resource"] = bool(buf.shift(1))
        if 0 != buf.shift(4):
            raise ValueError("RESERVED: buffer doesn't match value")
        result__el_tmp3["earfcn"] = buf.shift(16) + 0
        result__el_tmp3["pci"] = buf.shift(16) + 0
        result__el_tmp3["cell_id"] = buf.shift(16) + 0
        result__el_tmp3["tac"] = buf.shift(16) + 0
        result__el_tmp3["rsrp"] = buf.shift(8) + -140
        result__el_tmp3["rsrq"] = buf.shift(8) + -20
        result__el_tmp3["rssi"] = buf.shift(8) + -110
        result__el_tmp3["snr"] = buf.shift(8) + -20
        result__el_tmp3["band"] = buf.shift(8) + 0
        result__el_tmp3["ecl"] = buf.shift(8) + 0
        result__el_tmp3["tx_power"] = buf.shift(8) + -128
        result__el_tmp3["operation_mode"] = buf.shift(8) + 0
        result = SmpmUlDeviceGazFlow32BDailyData(**result__el_tmp3)
        return result

    def to_integration_data(self, received_at: datetime, device_tz: tzinfo, **kwargs: Any) -> List[IntegrationV0MessageData]:
        messages = []
        error_volume = 127  # by documentation
        v = int(self.direct_flow_volume_day_ago * 10)
        c = self.direct_flow_volume - self.direct_flow_volume_day_ago
        if v != 0 and v != error_volume and c > 0:
            messages.append(IntegrationV0MessageData(
                dt=datetime.combine(received_at.astimezone(device_tz).date(), time(0)).astimezone(timezone('UTC')),
                consumption=[
                    IntegrationV0MessageConsumption(
                        counter_type=CounterType.COMMON,
                        resource_type=ResourceType.COMMON,
                        channel=1,
                        value=Decimal(str(c)),
                        overloading_value=Decimal(str(4294967.295)),
                        journal_data_type=JournalDataType.END_OF_DAY if self.days_ago else JournalDataType.CURRENT,
                    ),
                ],
            ))
        return [
            *messages,
            IntegrationV0MessageData(
                dt=days_ago_calculation(received_at, device_tz, time(0), self.days_ago),
                sensors=[
                    IntegrationV0MessageSensor(
                        value=Decimal(str(self.battery_volts)),
                        sensor_type=SensorType.BATTERY,
                    ),
                    IntegrationV0MessageSensor(
                        value=Decimal(str(self.temperature)),
                        sensor_type=SensorType.TEMPERATURE,
                    ),
                ],
                consumption=[
                    IntegrationV0MessageConsumption(
                        counter_type=CounterType.COMMON,
                        resource_type=ResourceType.COMMON,
                        channel=1,
                        value=Decimal(str(self.direct_flow_volume)),
                        overloading_value=Decimal(str(4294967.295)),
                        journal_data_type=JournalDataType.END_OF_DAY if self.days_ago else JournalDataType.CURRENT,
                    ),
                ],
                events=[
                    *([IntegrationV0MessageEvent.RESET] if self.event_reset else []),
                    *([IntegrationV0MessageEvent.BATTERY_IS_LOW] if self.event_low_battery_level else []),
                    *([IntegrationV0MessageEvent.TEMPERATURE_LIMIT] if self.event_temperature_limits else []),
                    *([IntegrationV0MessageEvent.BATTERY_WARNING] if self.event_battery_warn else []),
                    *([IntegrationV0MessageEvent.ERROR_SYSTEM] if self.event_system_error else []),
                    *([IntegrationV0MessageEvent.FLOW_REVERSE] if self.event_flow_reverse else []),
                    *([IntegrationV0MessageEvent.FLOW_SPEED_OVER_LIMIT] if self.event_flow_speed_is_over_limit else []),
                    *([IntegrationV0MessageEvent.ERROR_SENSOR] if self.event_sensor_error else []),
                    *([IntegrationV0MessageEvent.ERROR_SENSOR_TEMPERATURE] if self.event_sensor_error_temperature else []),
                    *([IntegrationV0MessageEvent.CASE_WAS_OPENED] if self.event_case_was_opened else []),
                    *([IntegrationV0MessageEvent.CONTINUES_CONSUMPTION] if self.event_continuous_consumption else []),
                    *([IntegrationV0MessageEvent.NO_RESOURCE] if self.event_no_resource else []),
                ],
                generation=[
                    IntegrationV0MessageGeneration(
                        counter_type=CounterType.COMMON,
                        resource_type=ResourceType.COMMON,
                        value=Decimal(str(self.reverse_flow_volume)),
                        overloading_value=Decimal(str(40.95)),
                        journal_data_type=JournalDataType.END_OF_DAY if self.days_ago else JournalDataType.CURRENT,
                    ),
                ],
                clock=[IntegrationV0MessageClock(value=timestamp_calculation(received_at, self.timestamp_s, timedelta(seconds=67108863)))] if self.timestamp_s else [],
                connection_quality=IntegrationV0MessageConnectionQuality(
                    earfcn=self.earfcn,
                    pci=self.pci,
                    cell_id=self.cell_id,
                    tac=self.tac,
                    rsrp=self.rsrp,
                    rsrq=self.rsrq,
                    rssi=self.rssi,
                    snr=self.snr,
                    band=self.band,
                    ecl=self.ecl,
                    tx_power=self.tx_power,
                    operation_mode=self.operation_mode,
                )
            ),
        ]



from datetime import timedelta

from data_gateway_sdk.protocols.smpm.smpm_ul_device_energy_16b_daily import SmpmUlDeviceEnergy16BDailyData
from data_gateway_sdk.utils.buf_ref import BufRef


def test_smpm_ul_device_energy_16b_daily() -> None:
    case_serialized = bytes.fromhex("bbaaaaae545f55aaaa2abbaa6a80160b")
    assert SmpmUlDeviceEnergy16BDailyData(energy_consumed_active=1430869, energy_consumed_reactive=1398741, energy_generated_active=1398101, energy_generated_reactive=6990523, days_ago=timedelta(seconds=0.0), valid=False, error_measurement=True, error_low_voltage=False, error_internal_clock=True, error_flash=True, error_eeprom=False, error_radio=True, error_display=False, error_plc=False, error_reset=False, impact_power_lost=True, impact_magnet=True, impact_cleat_tamper=False, impact_body_tamper=True, impact_radio=False) == SmpmUlDeviceEnergy16BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BDailyData.serialize(SmpmUlDeviceEnergy16BDailyData(energy_consumed_active=1430869, energy_consumed_reactive=1398741, energy_generated_active=1398101, energy_generated_reactive=6990523, days_ago=timedelta(seconds=0.0), valid=False, error_measurement=True, error_low_voltage=False, error_internal_clock=True, error_flash=True, error_eeprom=False, error_radio=True, error_display=False, error_plc=False, error_reset=False, impact_power_lost=True, impact_magnet=True, impact_cleat_tamper=False, impact_body_tamper=True, impact_radio=False))  # noqa: E501
    case_serialized = bytes.fromhex("bb020000000000000000000000c0ff1f")  # overflow
    assert SmpmUlDeviceEnergy16BDailyData(energy_consumed_active=0, energy_consumed_reactive=0, energy_generated_active=0, energy_generated_reactive=0, days_ago=timedelta(seconds=0.0), valid=True, error_measurement=True, error_low_voltage=True, error_internal_clock=True, error_flash=True, error_eeprom=True, error_radio=True, error_display=True, error_plc=True, error_reset=True, impact_power_lost=True, impact_magnet=True, impact_cleat_tamper=True, impact_body_tamper=True, impact_radio=True) == SmpmUlDeviceEnergy16BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BDailyData.serialize(SmpmUlDeviceEnergy16BDailyData(days_ago=timedelta(seconds=0.0), energy_consumed_active=0, energy_consumed_reactive=0, energy_generated_active=0, energy_generated_reactive=0, error_display=True, error_eeprom=True, error_flash=True, error_internal_clock=True, error_low_voltage=True, error_measurement=True, error_plc=True, error_radio=True, error_reset=True, impact_body_tamper=True, impact_cleat_tamper=True, impact_magnet=True, impact_power_lost=True, impact_radio=True, valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("bb020000000000000000000000c0ff1f")  # overflow
    assert SmpmUlDeviceEnergy16BDailyData(energy_consumed_active=0, energy_consumed_reactive=0, energy_generated_active=0, energy_generated_reactive=0, days_ago=timedelta(seconds=0.0), valid=True, error_measurement=True, error_low_voltage=True, error_internal_clock=True, error_flash=True, error_eeprom=True, error_radio=True, error_display=True, error_plc=True, error_reset=True, impact_power_lost=True, impact_magnet=True, impact_cleat_tamper=True, impact_body_tamper=True, impact_radio=True) == SmpmUlDeviceEnergy16BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BDailyData.serialize(SmpmUlDeviceEnergy16BDailyData(days_ago=timedelta(seconds=0.0), energy_consumed_active=0, energy_consumed_reactive=0, energy_generated_active=0, energy_generated_reactive=16777216, error_display=True, error_eeprom=True, error_flash=True, error_internal_clock=True, error_low_voltage=True, error_measurement=True, error_plc=True, error_radio=True, error_reset=True, impact_body_tamper=True, impact_cleat_tamper=True, impact_magnet=True, impact_power_lost=True, impact_radio=True, valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("bb020000000000000000000000c0ff1f")  # overflow
    assert SmpmUlDeviceEnergy16BDailyData(energy_consumed_active=0, energy_consumed_reactive=0, energy_generated_active=0, energy_generated_reactive=0, days_ago=timedelta(seconds=0.0), valid=True, error_measurement=True, error_low_voltage=True, error_internal_clock=True, error_flash=True, error_eeprom=True, error_radio=True, error_display=True, error_plc=True, error_reset=True, impact_power_lost=True, impact_magnet=True, impact_cleat_tamper=True, impact_body_tamper=True, impact_radio=True) == SmpmUlDeviceEnergy16BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BDailyData.serialize(SmpmUlDeviceEnergy16BDailyData(days_ago=timedelta(seconds=0.0), energy_consumed_active=0, energy_consumed_reactive=0, energy_generated_active=16777216, energy_generated_reactive=0, error_display=True, error_eeprom=True, error_flash=True, error_internal_clock=True, error_low_voltage=True, error_measurement=True, error_plc=True, error_radio=True, error_reset=True, impact_body_tamper=True, impact_cleat_tamper=True, impact_magnet=True, impact_power_lost=True, impact_radio=True, valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("bb020000000000000000000000c0ff1f")  # overflow
    assert SmpmUlDeviceEnergy16BDailyData(energy_consumed_active=0, energy_consumed_reactive=0, energy_generated_active=0, energy_generated_reactive=0, days_ago=timedelta(seconds=0.0), valid=True, error_measurement=True, error_low_voltage=True, error_internal_clock=True, error_flash=True, error_eeprom=True, error_radio=True, error_display=True, error_plc=True, error_reset=True, impact_power_lost=True, impact_magnet=True, impact_cleat_tamper=True, impact_body_tamper=True, impact_radio=True) == SmpmUlDeviceEnergy16BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BDailyData.serialize(SmpmUlDeviceEnergy16BDailyData(days_ago=timedelta(seconds=0.0), energy_consumed_active=0, energy_consumed_reactive=0, energy_generated_active=16777216, energy_generated_reactive=16777216, error_display=True, error_eeprom=True, error_flash=True, error_internal_clock=True, error_low_voltage=True, error_measurement=True, error_plc=True, error_radio=True, error_reset=True, impact_body_tamper=True, impact_cleat_tamper=True, impact_magnet=True, impact_power_lost=True, impact_radio=True, valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("bb020000000000000000000000c0ff1f")  # overflow
    assert SmpmUlDeviceEnergy16BDailyData(energy_consumed_active=0, energy_consumed_reactive=0, energy_generated_active=0, energy_generated_reactive=0, days_ago=timedelta(seconds=0.0), valid=True, error_measurement=True, error_low_voltage=True, error_internal_clock=True, error_flash=True, error_eeprom=True, error_radio=True, error_display=True, error_plc=True, error_reset=True, impact_power_lost=True, impact_magnet=True, impact_cleat_tamper=True, impact_body_tamper=True, impact_radio=True) == SmpmUlDeviceEnergy16BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BDailyData.serialize(SmpmUlDeviceEnergy16BDailyData(days_ago=timedelta(seconds=0.0), energy_consumed_active=0, energy_consumed_reactive=16777216, energy_generated_active=0, energy_generated_reactive=0, error_display=True, error_eeprom=True, error_flash=True, error_internal_clock=True, error_low_voltage=True, error_measurement=True, error_plc=True, error_radio=True, error_reset=True, impact_body_tamper=True, impact_cleat_tamper=True, impact_magnet=True, impact_power_lost=True, impact_radio=True, valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("bb020000000000000000000000c0ff1f")  # overflow
    assert SmpmUlDeviceEnergy16BDailyData(energy_consumed_active=0, energy_consumed_reactive=0, energy_generated_active=0, energy_generated_reactive=0, days_ago=timedelta(seconds=0.0), valid=True, error_measurement=True, error_low_voltage=True, error_internal_clock=True, error_flash=True, error_eeprom=True, error_radio=True, error_display=True, error_plc=True, error_reset=True, impact_power_lost=True, impact_magnet=True, impact_cleat_tamper=True, impact_body_tamper=True, impact_radio=True) == SmpmUlDeviceEnergy16BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BDailyData.serialize(SmpmUlDeviceEnergy16BDailyData(days_ago=timedelta(seconds=0.0), energy_consumed_active=0, energy_consumed_reactive=16777216, energy_generated_active=0, energy_generated_reactive=16777216, error_display=True, error_eeprom=True, error_flash=True, error_internal_clock=True, error_low_voltage=True, error_measurement=True, error_plc=True, error_radio=True, error_reset=True, impact_body_tamper=True, impact_cleat_tamper=True, impact_magnet=True, impact_power_lost=True, impact_radio=True, valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("bb020000000000000000000000c0ff1f")  # overflow
    assert SmpmUlDeviceEnergy16BDailyData(energy_consumed_active=0, energy_consumed_reactive=0, energy_generated_active=0, energy_generated_reactive=0, days_ago=timedelta(seconds=0.0), valid=True, error_measurement=True, error_low_voltage=True, error_internal_clock=True, error_flash=True, error_eeprom=True, error_radio=True, error_display=True, error_plc=True, error_reset=True, impact_power_lost=True, impact_magnet=True, impact_cleat_tamper=True, impact_body_tamper=True, impact_radio=True) == SmpmUlDeviceEnergy16BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BDailyData.serialize(SmpmUlDeviceEnergy16BDailyData(days_ago=timedelta(seconds=0.0), energy_consumed_active=0, energy_consumed_reactive=16777216, energy_generated_active=16777216, energy_generated_reactive=0, error_display=True, error_eeprom=True, error_flash=True, error_internal_clock=True, error_low_voltage=True, error_measurement=True, error_plc=True, error_radio=True, error_reset=True, impact_body_tamper=True, impact_cleat_tamper=True, impact_magnet=True, impact_power_lost=True, impact_radio=True, valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("bb020000000000000000000000c0ff1f")  # overflow
    assert SmpmUlDeviceEnergy16BDailyData(energy_consumed_active=0, energy_consumed_reactive=0, energy_generated_active=0, energy_generated_reactive=0, days_ago=timedelta(seconds=0.0), valid=True, error_measurement=True, error_low_voltage=True, error_internal_clock=True, error_flash=True, error_eeprom=True, error_radio=True, error_display=True, error_plc=True, error_reset=True, impact_power_lost=True, impact_magnet=True, impact_cleat_tamper=True, impact_body_tamper=True, impact_radio=True) == SmpmUlDeviceEnergy16BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BDailyData.serialize(SmpmUlDeviceEnergy16BDailyData(days_ago=timedelta(seconds=0.0), energy_consumed_active=0, energy_consumed_reactive=16777216, energy_generated_active=16777216, energy_generated_reactive=16777216, error_display=True, error_eeprom=True, error_flash=True, error_internal_clock=True, error_low_voltage=True, error_measurement=True, error_plc=True, error_radio=True, error_reset=True, impact_body_tamper=True, impact_cleat_tamper=True, impact_magnet=True, impact_power_lost=True, impact_radio=True, valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("bb020000000000000000000000c0ff1f")  # overflow
    assert SmpmUlDeviceEnergy16BDailyData(energy_consumed_active=0, energy_consumed_reactive=0, energy_generated_active=0, energy_generated_reactive=0, days_ago=timedelta(seconds=0.0), valid=True, error_measurement=True, error_low_voltage=True, error_internal_clock=True, error_flash=True, error_eeprom=True, error_radio=True, error_display=True, error_plc=True, error_reset=True, impact_power_lost=True, impact_magnet=True, impact_cleat_tamper=True, impact_body_tamper=True, impact_radio=True) == SmpmUlDeviceEnergy16BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BDailyData.serialize(SmpmUlDeviceEnergy16BDailyData(days_ago=timedelta(seconds=0.0), energy_consumed_active=16777216, energy_consumed_reactive=0, energy_generated_active=0, energy_generated_reactive=0, error_display=True, error_eeprom=True, error_flash=True, error_internal_clock=True, error_low_voltage=True, error_measurement=True, error_plc=True, error_radio=True, error_reset=True, impact_body_tamper=True, impact_cleat_tamper=True, impact_magnet=True, impact_power_lost=True, impact_radio=True, valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("bb020000000000000000000000c0ff1f")  # overflow
    assert SmpmUlDeviceEnergy16BDailyData(energy_consumed_active=0, energy_consumed_reactive=0, energy_generated_active=0, energy_generated_reactive=0, days_ago=timedelta(seconds=0.0), valid=True, error_measurement=True, error_low_voltage=True, error_internal_clock=True, error_flash=True, error_eeprom=True, error_radio=True, error_display=True, error_plc=True, error_reset=True, impact_power_lost=True, impact_magnet=True, impact_cleat_tamper=True, impact_body_tamper=True, impact_radio=True) == SmpmUlDeviceEnergy16BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BDailyData.serialize(SmpmUlDeviceEnergy16BDailyData(days_ago=timedelta(seconds=0.0), energy_consumed_active=16777216, energy_consumed_reactive=0, energy_generated_active=0, energy_generated_reactive=16777216, error_display=True, error_eeprom=True, error_flash=True, error_internal_clock=True, error_low_voltage=True, error_measurement=True, error_plc=True, error_radio=True, error_reset=True, impact_body_tamper=True, impact_cleat_tamper=True, impact_magnet=True, impact_power_lost=True, impact_radio=True, valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("bb020000000000000000000000c0ff1f")  # overflow
    assert SmpmUlDeviceEnergy16BDailyData(energy_consumed_active=0, energy_consumed_reactive=0, energy_generated_active=0, energy_generated_reactive=0, days_ago=timedelta(seconds=0.0), valid=True, error_measurement=True, error_low_voltage=True, error_internal_clock=True, error_flash=True, error_eeprom=True, error_radio=True, error_display=True, error_plc=True, error_reset=True, impact_power_lost=True, impact_magnet=True, impact_cleat_tamper=True, impact_body_tamper=True, impact_radio=True) == SmpmUlDeviceEnergy16BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BDailyData.serialize(SmpmUlDeviceEnergy16BDailyData(days_ago=timedelta(seconds=0.0), energy_consumed_active=16777216, energy_consumed_reactive=0, energy_generated_active=16777216, energy_generated_reactive=0, error_display=True, error_eeprom=True, error_flash=True, error_internal_clock=True, error_low_voltage=True, error_measurement=True, error_plc=True, error_radio=True, error_reset=True, impact_body_tamper=True, impact_cleat_tamper=True, impact_magnet=True, impact_power_lost=True, impact_radio=True, valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("bb020000000000000000000000c0ff1f")  # overflow
    assert SmpmUlDeviceEnergy16BDailyData(energy_consumed_active=0, energy_consumed_reactive=0, energy_generated_active=0, energy_generated_reactive=0, days_ago=timedelta(seconds=0.0), valid=True, error_measurement=True, error_low_voltage=True, error_internal_clock=True, error_flash=True, error_eeprom=True, error_radio=True, error_display=True, error_plc=True, error_reset=True, impact_power_lost=True, impact_magnet=True, impact_cleat_tamper=True, impact_body_tamper=True, impact_radio=True) == SmpmUlDeviceEnergy16BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BDailyData.serialize(SmpmUlDeviceEnergy16BDailyData(days_ago=timedelta(seconds=0.0), energy_consumed_active=16777216, energy_consumed_reactive=0, energy_generated_active=16777216, energy_generated_reactive=16777216, error_display=True, error_eeprom=True, error_flash=True, error_internal_clock=True, error_low_voltage=True, error_measurement=True, error_plc=True, error_radio=True, error_reset=True, impact_body_tamper=True, impact_cleat_tamper=True, impact_magnet=True, impact_power_lost=True, impact_radio=True, valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("bb020000000000000000000000c0ff1f")  # overflow
    assert SmpmUlDeviceEnergy16BDailyData(energy_consumed_active=0, energy_consumed_reactive=0, energy_generated_active=0, energy_generated_reactive=0, days_ago=timedelta(seconds=0.0), valid=True, error_measurement=True, error_low_voltage=True, error_internal_clock=True, error_flash=True, error_eeprom=True, error_radio=True, error_display=True, error_plc=True, error_reset=True, impact_power_lost=True, impact_magnet=True, impact_cleat_tamper=True, impact_body_tamper=True, impact_radio=True) == SmpmUlDeviceEnergy16BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BDailyData.serialize(SmpmUlDeviceEnergy16BDailyData(days_ago=timedelta(seconds=0.0), energy_consumed_active=16777216, energy_consumed_reactive=16777216, energy_generated_active=0, energy_generated_reactive=0, error_display=True, error_eeprom=True, error_flash=True, error_internal_clock=True, error_low_voltage=True, error_measurement=True, error_plc=True, error_radio=True, error_reset=True, impact_body_tamper=True, impact_cleat_tamper=True, impact_magnet=True, impact_power_lost=True, impact_radio=True, valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("bb020000000000000000000000c0ff1f")  # overflow
    assert SmpmUlDeviceEnergy16BDailyData(energy_consumed_active=0, energy_consumed_reactive=0, energy_generated_active=0, energy_generated_reactive=0, days_ago=timedelta(seconds=0.0), valid=True, error_measurement=True, error_low_voltage=True, error_internal_clock=True, error_flash=True, error_eeprom=True, error_radio=True, error_display=True, error_plc=True, error_reset=True, impact_power_lost=True, impact_magnet=True, impact_cleat_tamper=True, impact_body_tamper=True, impact_radio=True) == SmpmUlDeviceEnergy16BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BDailyData.serialize(SmpmUlDeviceEnergy16BDailyData(days_ago=timedelta(seconds=0.0), energy_consumed_active=16777216, energy_consumed_reactive=16777216, energy_generated_active=0, energy_generated_reactive=16777216, error_display=True, error_eeprom=True, error_flash=True, error_internal_clock=True, error_low_voltage=True, error_measurement=True, error_plc=True, error_radio=True, error_reset=True, impact_body_tamper=True, impact_cleat_tamper=True, impact_magnet=True, impact_power_lost=True, impact_radio=True, valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("bb020000000000000000000000c0ff1f")  # overflow
    assert SmpmUlDeviceEnergy16BDailyData(energy_consumed_active=0, energy_consumed_reactive=0, energy_generated_active=0, energy_generated_reactive=0, days_ago=timedelta(seconds=0.0), valid=True, error_measurement=True, error_low_voltage=True, error_internal_clock=True, error_flash=True, error_eeprom=True, error_radio=True, error_display=True, error_plc=True, error_reset=True, impact_power_lost=True, impact_magnet=True, impact_cleat_tamper=True, impact_body_tamper=True, impact_radio=True) == SmpmUlDeviceEnergy16BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BDailyData.serialize(SmpmUlDeviceEnergy16BDailyData(days_ago=timedelta(seconds=0.0), energy_consumed_active=16777216, energy_consumed_reactive=16777216, energy_generated_active=16777216, energy_generated_reactive=0, error_display=True, error_eeprom=True, error_flash=True, error_internal_clock=True, error_low_voltage=True, error_measurement=True, error_plc=True, error_radio=True, error_reset=True, impact_body_tamper=True, impact_cleat_tamper=True, impact_magnet=True, impact_power_lost=True, impact_radio=True, valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("bb020000000000000000000000c0ff1f")  # overflow
    assert SmpmUlDeviceEnergy16BDailyData(energy_consumed_active=0, energy_consumed_reactive=0, energy_generated_active=0, energy_generated_reactive=0, days_ago=timedelta(seconds=0.0), valid=True, error_measurement=True, error_low_voltage=True, error_internal_clock=True, error_flash=True, error_eeprom=True, error_radio=True, error_display=True, error_plc=True, error_reset=True, impact_power_lost=True, impact_magnet=True, impact_cleat_tamper=True, impact_body_tamper=True, impact_radio=True) == SmpmUlDeviceEnergy16BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BDailyData.serialize(SmpmUlDeviceEnergy16BDailyData(days_ago=timedelta(seconds=0.0), energy_consumed_active=16777216, energy_consumed_reactive=16777216, energy_generated_active=16777216, energy_generated_reactive=16777216, error_display=True, error_eeprom=True, error_flash=True, error_internal_clock=True, error_low_voltage=True, error_measurement=True, error_plc=True, error_radio=True, error_reset=True, impact_body_tamper=True, impact_cleat_tamper=True, impact_magnet=True, impact_power_lost=True, impact_radio=True, valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("bb020000000000000000000080ffff1f")  # overflow
    assert SmpmUlDeviceEnergy16BDailyData(energy_consumed_active=0, energy_consumed_reactive=0, energy_generated_active=0, energy_generated_reactive=0, days_ago=timedelta(seconds=10972800.0), valid=True, error_measurement=True, error_low_voltage=True, error_internal_clock=True, error_flash=True, error_eeprom=True, error_radio=True, error_display=True, error_plc=True, error_reset=True, impact_power_lost=True, impact_magnet=True, impact_cleat_tamper=True, impact_body_tamper=True, impact_radio=True) == SmpmUlDeviceEnergy16BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BDailyData.serialize(SmpmUlDeviceEnergy16BDailyData(days_ago=timedelta(seconds=10972800.0), energy_consumed_active=0, energy_consumed_reactive=0, energy_generated_active=0, energy_generated_reactive=0, error_display=True, error_eeprom=True, error_flash=True, error_internal_clock=True, error_low_voltage=True, error_measurement=True, error_plc=True, error_radio=True, error_reset=True, impact_body_tamper=True, impact_cleat_tamper=True, impact_magnet=True, impact_power_lost=True, impact_radio=True, valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("bb020000000000000000000080ffff1f")  # overflow
    assert SmpmUlDeviceEnergy16BDailyData(energy_consumed_active=0, energy_consumed_reactive=0, energy_generated_active=0, energy_generated_reactive=0, days_ago=timedelta(seconds=10972800.0), valid=True, error_measurement=True, error_low_voltage=True, error_internal_clock=True, error_flash=True, error_eeprom=True, error_radio=True, error_display=True, error_plc=True, error_reset=True, impact_power_lost=True, impact_magnet=True, impact_cleat_tamper=True, impact_body_tamper=True, impact_radio=True) == SmpmUlDeviceEnergy16BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BDailyData.serialize(SmpmUlDeviceEnergy16BDailyData(days_ago=timedelta(seconds=10972800.0), energy_consumed_active=0, energy_consumed_reactive=0, energy_generated_active=0, energy_generated_reactive=16777216, error_display=True, error_eeprom=True, error_flash=True, error_internal_clock=True, error_low_voltage=True, error_measurement=True, error_plc=True, error_radio=True, error_reset=True, impact_body_tamper=True, impact_cleat_tamper=True, impact_magnet=True, impact_power_lost=True, impact_radio=True, valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("bb020000000000000000000080ffff1f")  # overflow
    assert SmpmUlDeviceEnergy16BDailyData(energy_consumed_active=0, energy_consumed_reactive=0, energy_generated_active=0, energy_generated_reactive=0, days_ago=timedelta(seconds=10972800.0), valid=True, error_measurement=True, error_low_voltage=True, error_internal_clock=True, error_flash=True, error_eeprom=True, error_radio=True, error_display=True, error_plc=True, error_reset=True, impact_power_lost=True, impact_magnet=True, impact_cleat_tamper=True, impact_body_tamper=True, impact_radio=True) == SmpmUlDeviceEnergy16BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BDailyData.serialize(SmpmUlDeviceEnergy16BDailyData(days_ago=timedelta(seconds=10972800.0), energy_consumed_active=0, energy_consumed_reactive=0, energy_generated_active=16777216, energy_generated_reactive=0, error_display=True, error_eeprom=True, error_flash=True, error_internal_clock=True, error_low_voltage=True, error_measurement=True, error_plc=True, error_radio=True, error_reset=True, impact_body_tamper=True, impact_cleat_tamper=True, impact_magnet=True, impact_power_lost=True, impact_radio=True, valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("bb020000000000000000000080ffff1f")  # overflow
    assert SmpmUlDeviceEnergy16BDailyData(energy_consumed_active=0, energy_consumed_reactive=0, energy_generated_active=0, energy_generated_reactive=0, days_ago=timedelta(seconds=10972800.0), valid=True, error_measurement=True, error_low_voltage=True, error_internal_clock=True, error_flash=True, error_eeprom=True, error_radio=True, error_display=True, error_plc=True, error_reset=True, impact_power_lost=True, impact_magnet=True, impact_cleat_tamper=True, impact_body_tamper=True, impact_radio=True) == SmpmUlDeviceEnergy16BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BDailyData.serialize(SmpmUlDeviceEnergy16BDailyData(days_ago=timedelta(seconds=10972800.0), energy_consumed_active=0, energy_consumed_reactive=0, energy_generated_active=16777216, energy_generated_reactive=16777216, error_display=True, error_eeprom=True, error_flash=True, error_internal_clock=True, error_low_voltage=True, error_measurement=True, error_plc=True, error_radio=True, error_reset=True, impact_body_tamper=True, impact_cleat_tamper=True, impact_magnet=True, impact_power_lost=True, impact_radio=True, valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("bb020000000000000000000080ffff1f")  # overflow
    assert SmpmUlDeviceEnergy16BDailyData(energy_consumed_active=0, energy_consumed_reactive=0, energy_generated_active=0, energy_generated_reactive=0, days_ago=timedelta(seconds=10972800.0), valid=True, error_measurement=True, error_low_voltage=True, error_internal_clock=True, error_flash=True, error_eeprom=True, error_radio=True, error_display=True, error_plc=True, error_reset=True, impact_power_lost=True, impact_magnet=True, impact_cleat_tamper=True, impact_body_tamper=True, impact_radio=True) == SmpmUlDeviceEnergy16BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BDailyData.serialize(SmpmUlDeviceEnergy16BDailyData(days_ago=timedelta(seconds=10972800.0), energy_consumed_active=0, energy_consumed_reactive=16777216, energy_generated_active=0, energy_generated_reactive=0, error_display=True, error_eeprom=True, error_flash=True, error_internal_clock=True, error_low_voltage=True, error_measurement=True, error_plc=True, error_radio=True, error_reset=True, impact_body_tamper=True, impact_cleat_tamper=True, impact_magnet=True, impact_power_lost=True, impact_radio=True, valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("bb020000000000000000000080ffff1f")  # overflow
    assert SmpmUlDeviceEnergy16BDailyData(energy_consumed_active=0, energy_consumed_reactive=0, energy_generated_active=0, energy_generated_reactive=0, days_ago=timedelta(seconds=10972800.0), valid=True, error_measurement=True, error_low_voltage=True, error_internal_clock=True, error_flash=True, error_eeprom=True, error_radio=True, error_display=True, error_plc=True, error_reset=True, impact_power_lost=True, impact_magnet=True, impact_cleat_tamper=True, impact_body_tamper=True, impact_radio=True) == SmpmUlDeviceEnergy16BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BDailyData.serialize(SmpmUlDeviceEnergy16BDailyData(days_ago=timedelta(seconds=10972800.0), energy_consumed_active=0, energy_consumed_reactive=16777216, energy_generated_active=0, energy_generated_reactive=16777216, error_display=True, error_eeprom=True, error_flash=True, error_internal_clock=True, error_low_voltage=True, error_measurement=True, error_plc=True, error_radio=True, error_reset=True, impact_body_tamper=True, impact_cleat_tamper=True, impact_magnet=True, impact_power_lost=True, impact_radio=True, valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("bb020000000000000000000080ffff1f")  # overflow
    assert SmpmUlDeviceEnergy16BDailyData(energy_consumed_active=0, energy_consumed_reactive=0, energy_generated_active=0, energy_generated_reactive=0, days_ago=timedelta(seconds=10972800.0), valid=True, error_measurement=True, error_low_voltage=True, error_internal_clock=True, error_flash=True, error_eeprom=True, error_radio=True, error_display=True, error_plc=True, error_reset=True, impact_power_lost=True, impact_magnet=True, impact_cleat_tamper=True, impact_body_tamper=True, impact_radio=True) == SmpmUlDeviceEnergy16BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BDailyData.serialize(SmpmUlDeviceEnergy16BDailyData(days_ago=timedelta(seconds=10972800.0), energy_consumed_active=0, energy_consumed_reactive=16777216, energy_generated_active=16777216, energy_generated_reactive=0, error_display=True, error_eeprom=True, error_flash=True, error_internal_clock=True, error_low_voltage=True, error_measurement=True, error_plc=True, error_radio=True, error_reset=True, impact_body_tamper=True, impact_cleat_tamper=True, impact_magnet=True, impact_power_lost=True, impact_radio=True, valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("bb020000000000000000000080ffff1f")  # overflow
    assert SmpmUlDeviceEnergy16BDailyData(energy_consumed_active=0, energy_consumed_reactive=0, energy_generated_active=0, energy_generated_reactive=0, days_ago=timedelta(seconds=10972800.0), valid=True, error_measurement=True, error_low_voltage=True, error_internal_clock=True, error_flash=True, error_eeprom=True, error_radio=True, error_display=True, error_plc=True, error_reset=True, impact_power_lost=True, impact_magnet=True, impact_cleat_tamper=True, impact_body_tamper=True, impact_radio=True) == SmpmUlDeviceEnergy16BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BDailyData.serialize(SmpmUlDeviceEnergy16BDailyData(days_ago=timedelta(seconds=10972800.0), energy_consumed_active=0, energy_consumed_reactive=16777216, energy_generated_active=16777216, energy_generated_reactive=16777216, error_display=True, error_eeprom=True, error_flash=True, error_internal_clock=True, error_low_voltage=True, error_measurement=True, error_plc=True, error_radio=True, error_reset=True, impact_body_tamper=True, impact_cleat_tamper=True, impact_magnet=True, impact_power_lost=True, impact_radio=True, valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("bb020000000000000000000080ffff1f")  # overflow
    assert SmpmUlDeviceEnergy16BDailyData(energy_consumed_active=0, energy_consumed_reactive=0, energy_generated_active=0, energy_generated_reactive=0, days_ago=timedelta(seconds=10972800.0), valid=True, error_measurement=True, error_low_voltage=True, error_internal_clock=True, error_flash=True, error_eeprom=True, error_radio=True, error_display=True, error_plc=True, error_reset=True, impact_power_lost=True, impact_magnet=True, impact_cleat_tamper=True, impact_body_tamper=True, impact_radio=True) == SmpmUlDeviceEnergy16BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BDailyData.serialize(SmpmUlDeviceEnergy16BDailyData(days_ago=timedelta(seconds=10972800.0), energy_consumed_active=16777216, energy_consumed_reactive=0, energy_generated_active=0, energy_generated_reactive=0, error_display=True, error_eeprom=True, error_flash=True, error_internal_clock=True, error_low_voltage=True, error_measurement=True, error_plc=True, error_radio=True, error_reset=True, impact_body_tamper=True, impact_cleat_tamper=True, impact_magnet=True, impact_power_lost=True, impact_radio=True, valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("bb020000000000000000000080ffff1f")  # overflow
    assert SmpmUlDeviceEnergy16BDailyData(energy_consumed_active=0, energy_consumed_reactive=0, energy_generated_active=0, energy_generated_reactive=0, days_ago=timedelta(seconds=10972800.0), valid=True, error_measurement=True, error_low_voltage=True, error_internal_clock=True, error_flash=True, error_eeprom=True, error_radio=True, error_display=True, error_plc=True, error_reset=True, impact_power_lost=True, impact_magnet=True, impact_cleat_tamper=True, impact_body_tamper=True, impact_radio=True) == SmpmUlDeviceEnergy16BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BDailyData.serialize(SmpmUlDeviceEnergy16BDailyData(days_ago=timedelta(seconds=10972800.0), energy_consumed_active=16777216, energy_consumed_reactive=0, energy_generated_active=0, energy_generated_reactive=16777216, error_display=True, error_eeprom=True, error_flash=True, error_internal_clock=True, error_low_voltage=True, error_measurement=True, error_plc=True, error_radio=True, error_reset=True, impact_body_tamper=True, impact_cleat_tamper=True, impact_magnet=True, impact_power_lost=True, impact_radio=True, valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("bb020000000000000000000080ffff1f")  # overflow
    assert SmpmUlDeviceEnergy16BDailyData(energy_consumed_active=0, energy_consumed_reactive=0, energy_generated_active=0, energy_generated_reactive=0, days_ago=timedelta(seconds=10972800.0), valid=True, error_measurement=True, error_low_voltage=True, error_internal_clock=True, error_flash=True, error_eeprom=True, error_radio=True, error_display=True, error_plc=True, error_reset=True, impact_power_lost=True, impact_magnet=True, impact_cleat_tamper=True, impact_body_tamper=True, impact_radio=True) == SmpmUlDeviceEnergy16BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BDailyData.serialize(SmpmUlDeviceEnergy16BDailyData(days_ago=timedelta(seconds=10972800.0), energy_consumed_active=16777216, energy_consumed_reactive=0, energy_generated_active=16777216, energy_generated_reactive=0, error_display=True, error_eeprom=True, error_flash=True, error_internal_clock=True, error_low_voltage=True, error_measurement=True, error_plc=True, error_radio=True, error_reset=True, impact_body_tamper=True, impact_cleat_tamper=True, impact_magnet=True, impact_power_lost=True, impact_radio=True, valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("bb020000000000000000000080ffff1f")  # overflow
    assert SmpmUlDeviceEnergy16BDailyData(energy_consumed_active=0, energy_consumed_reactive=0, energy_generated_active=0, energy_generated_reactive=0, days_ago=timedelta(seconds=10972800.0), valid=True, error_measurement=True, error_low_voltage=True, error_internal_clock=True, error_flash=True, error_eeprom=True, error_radio=True, error_display=True, error_plc=True, error_reset=True, impact_power_lost=True, impact_magnet=True, impact_cleat_tamper=True, impact_body_tamper=True, impact_radio=True) == SmpmUlDeviceEnergy16BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BDailyData.serialize(SmpmUlDeviceEnergy16BDailyData(days_ago=timedelta(seconds=10972800.0), energy_consumed_active=16777216, energy_consumed_reactive=0, energy_generated_active=16777216, energy_generated_reactive=16777216, error_display=True, error_eeprom=True, error_flash=True, error_internal_clock=True, error_low_voltage=True, error_measurement=True, error_plc=True, error_radio=True, error_reset=True, impact_body_tamper=True, impact_cleat_tamper=True, impact_magnet=True, impact_power_lost=True, impact_radio=True, valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("bb020000000000000000000080ffff1f")  # overflow
    assert SmpmUlDeviceEnergy16BDailyData(energy_consumed_active=0, energy_consumed_reactive=0, energy_generated_active=0, energy_generated_reactive=0, days_ago=timedelta(seconds=10972800.0), valid=True, error_measurement=True, error_low_voltage=True, error_internal_clock=True, error_flash=True, error_eeprom=True, error_radio=True, error_display=True, error_plc=True, error_reset=True, impact_power_lost=True, impact_magnet=True, impact_cleat_tamper=True, impact_body_tamper=True, impact_radio=True) == SmpmUlDeviceEnergy16BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BDailyData.serialize(SmpmUlDeviceEnergy16BDailyData(days_ago=timedelta(seconds=10972800.0), energy_consumed_active=16777216, energy_consumed_reactive=16777216, energy_generated_active=0, energy_generated_reactive=0, error_display=True, error_eeprom=True, error_flash=True, error_internal_clock=True, error_low_voltage=True, error_measurement=True, error_plc=True, error_radio=True, error_reset=True, impact_body_tamper=True, impact_cleat_tamper=True, impact_magnet=True, impact_power_lost=True, impact_radio=True, valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("bb020000000000000000000080ffff1f")  # overflow
    assert SmpmUlDeviceEnergy16BDailyData(energy_consumed_active=0, energy_consumed_reactive=0, energy_generated_active=0, energy_generated_reactive=0, days_ago=timedelta(seconds=10972800.0), valid=True, error_measurement=True, error_low_voltage=True, error_internal_clock=True, error_flash=True, error_eeprom=True, error_radio=True, error_display=True, error_plc=True, error_reset=True, impact_power_lost=True, impact_magnet=True, impact_cleat_tamper=True, impact_body_tamper=True, impact_radio=True) == SmpmUlDeviceEnergy16BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BDailyData.serialize(SmpmUlDeviceEnergy16BDailyData(days_ago=timedelta(seconds=10972800.0), energy_consumed_active=16777216, energy_consumed_reactive=16777216, energy_generated_active=0, energy_generated_reactive=16777216, error_display=True, error_eeprom=True, error_flash=True, error_internal_clock=True, error_low_voltage=True, error_measurement=True, error_plc=True, error_radio=True, error_reset=True, impact_body_tamper=True, impact_cleat_tamper=True, impact_magnet=True, impact_power_lost=True, impact_radio=True, valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("bb020000000000000000000080ffff1f")  # overflow
    assert SmpmUlDeviceEnergy16BDailyData(energy_consumed_active=0, energy_consumed_reactive=0, energy_generated_active=0, energy_generated_reactive=0, days_ago=timedelta(seconds=10972800.0), valid=True, error_measurement=True, error_low_voltage=True, error_internal_clock=True, error_flash=True, error_eeprom=True, error_radio=True, error_display=True, error_plc=True, error_reset=True, impact_power_lost=True, impact_magnet=True, impact_cleat_tamper=True, impact_body_tamper=True, impact_radio=True) == SmpmUlDeviceEnergy16BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BDailyData.serialize(SmpmUlDeviceEnergy16BDailyData(days_ago=timedelta(seconds=10972800.0), energy_consumed_active=16777216, energy_consumed_reactive=16777216, energy_generated_active=16777216, energy_generated_reactive=0, error_display=True, error_eeprom=True, error_flash=True, error_internal_clock=True, error_low_voltage=True, error_measurement=True, error_plc=True, error_radio=True, error_reset=True, impact_body_tamper=True, impact_cleat_tamper=True, impact_magnet=True, impact_power_lost=True, impact_radio=True, valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("bb020000000000000000000080ffff1f")  # overflow
    assert SmpmUlDeviceEnergy16BDailyData(energy_consumed_active=0, energy_consumed_reactive=0, energy_generated_active=0, energy_generated_reactive=0, days_ago=timedelta(seconds=10972800.0), valid=True, error_measurement=True, error_low_voltage=True, error_internal_clock=True, error_flash=True, error_eeprom=True, error_radio=True, error_display=True, error_plc=True, error_reset=True, impact_power_lost=True, impact_magnet=True, impact_cleat_tamper=True, impact_body_tamper=True, impact_radio=True) == SmpmUlDeviceEnergy16BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BDailyData.serialize(SmpmUlDeviceEnergy16BDailyData(days_ago=timedelta(seconds=10972800.0), energy_consumed_active=16777216, energy_consumed_reactive=16777216, energy_generated_active=16777216, energy_generated_reactive=16777216, error_display=True, error_eeprom=True, error_flash=True, error_internal_clock=True, error_low_voltage=True, error_measurement=True, error_plc=True, error_radio=True, error_reset=True, impact_body_tamper=True, impact_cleat_tamper=True, impact_magnet=True, impact_power_lost=True, impact_radio=True, valid=True))  # noqa: E501



from enum import Enum, unique
from typing import Dict, Any

from data_gateway_sdk.errors import DataGatewayBSProtocolParsingError
from data_gateway_sdk.protocols.nero_bs_packet.http_nero_bs_packet import HttpV0NeroBsPacket
from data_gateway_sdk.protocols.nero_bs_packet.nero_bs_packet import NeroBsPacket


@unique
class NeroBsProtocolType(Enum):  # DO NOT CHANGE VALUE because it should be compatible with data_aggregator_sdk db
    NERO_BS_HTTP_V0 = 'NERO_BS_HTTP_V0'

    def __repr__(self) -> str:
        return f'{type(self).__name__}.{self.name}'

    def parse(self, payload: Dict[str, Any], **kwargs: Any) -> NeroBsPacket:
        assert isinstance(payload, dict), f'payload must be type of {dict}, got type {type(payload)}'
        try:
            return BS_PROTOCOLS_MAP[self].parse(payload, **kwargs)
        except DataGatewayBSProtocolParsingError:
            raise
        except Exception as e:  # noqa: B902
            raise DataGatewayBSProtocolParsingError(f'{self.value}: invalid payload', e)


BS_PROTOCOLS_MAP = {
    NeroBsProtocolType.NERO_BS_HTTP_V0: HttpV0NeroBsPacket,
}

assert set(BS_PROTOCOLS_MAP.keys()) == set(NeroBsProtocolType)



# from datetime import timedelta
#
# from data_gateway_sdk.protocols.smpm.smpm_ul_device_energy_16b_profile_8h_pqs import SmpmUlDeviceEnergy16BProfile8HPqsData, SmpmUlDeviceEnergy16bProfile8hPQSIds
# from data_gateway_sdk.utils.buf_ref import BufRef
#
#
# def test_smpm_ul_device_energy_16b_profile_8h_pqs() -> None:
#     case_serialized = bytes.fromhex("b10c008000086000042880010e200300")
#     assert SmpmUlDeviceEnergy16BProfile8HPqsData(packet_type_id=SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_1_S__AVG, days_ago=timedelta(seconds=0.0), profile=(0, 1, 2, 3, 4, 5, 6, 7), point_factor=5.0) == SmpmUlDeviceEnergy16BProfile8HPqsData.parse(BufRef(case_serialized))  # noqa: E501
#     assert case_serialized == SmpmUlDeviceEnergy16BProfile8HPqsData.serialize(SmpmUlDeviceEnergy16BProfile8HPqsData(packet_type_id=SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_1_S__AVG, days_ago=timedelta(seconds=0.0), profile=(0, 1, 2, 3, 4, 5, 6, 7), point_factor=5.0))  # noqa: E501
#     case_serialized = bytes.fromhex("b10c008000086000042880010e000000")
#     assert SmpmUlDeviceEnergy16BProfile8HPqsData(packet_type_id=SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_1_S__AVG, days_ago=timedelta(seconds=0.0), profile=(0, 1, 2, 3, 4, 5, 6, 7), point_factor=0.0) == SmpmUlDeviceEnergy16BProfile8HPqsData.parse(BufRef(case_serialized))  # noqa: E501
#     assert case_serialized == SmpmUlDeviceEnergy16BProfile8HPqsData.serialize(SmpmUlDeviceEnergy16BProfile8HPqsData(days_ago=timedelta(seconds=0.0), packet_type_id=SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_1_S__AVG, point_factor=0.0, profile=(0, 1, 2, 3, 4, 5, 6, 7)))  # noqa: E501
#     case_serialized = bytes.fromhex("b10c008000086000042880010ef0ff00")
#     assert SmpmUlDeviceEnergy16BProfile8HPqsData(packet_type_id=SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_1_S__AVG, days_ago=timedelta(seconds=0.0), profile=(0, 1, 2, 3, 4, 5, 6, 7), point_factor=409.5) == SmpmUlDeviceEnergy16BProfile8HPqsData.parse(BufRef(case_serialized))  # noqa: E501
#     assert case_serialized == SmpmUlDeviceEnergy16BProfile8HPqsData.serialize(SmpmUlDeviceEnergy16BProfile8HPqsData(days_ago=timedelta(seconds=0.0), packet_type_id=SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_1_S__AVG, point_factor=409.5, profile=(0, 1, 2, 3, 4, 5, 6, 7)))  # noqa: E501
#     case_serialized = bytes.fromhex("9c0d008000086000042880010e000000")
#     assert SmpmUlDeviceEnergy16BProfile8HPqsData(packet_type_id=SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_3_Q__MAX__PHASE_C, days_ago=timedelta(seconds=0.0), profile=(0, 1, 2, 3, 4, 5, 6, 7), point_factor=0.0) == SmpmUlDeviceEnergy16BProfile8HPqsData.parse(BufRef(case_serialized))  # noqa: E501
#     assert case_serialized == SmpmUlDeviceEnergy16BProfile8HPqsData.serialize(SmpmUlDeviceEnergy16BProfile8HPqsData(days_ago=timedelta(seconds=0.0), packet_type_id=SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_3_Q__MAX__PHASE_C, point_factor=0.0, profile=(0, 1, 2, 3, 4, 5, 6, 7)))  # noqa: E501
#     case_serialized = bytes.fromhex("9c0d008000086000042880010ef0ff00")
#     assert SmpmUlDeviceEnergy16BProfile8HPqsData(packet_type_id=SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_3_Q__MAX__PHASE_C, days_ago=timedelta(seconds=0.0), profile=(0, 1, 2, 3, 4, 5, 6, 7), point_factor=409.5) == SmpmUlDeviceEnergy16BProfile8HPqsData.parse(BufRef(case_serialized))  # noqa: E501
#     assert case_serialized == SmpmUlDeviceEnergy16BProfile8HPqsData.serialize(SmpmUlDeviceEnergy16BProfile8HPqsData(days_ago=timedelta(seconds=0.0), packet_type_id=SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_3_Q__MAX__PHASE_C, point_factor=409.5, profile=(0, 1, 2, 3, 4, 5, 6, 7)))  # noqa: E501
#     case_serialized = bytes.fromhex("b1cc0f8000086000042880010e000000")
#     assert SmpmUlDeviceEnergy16BProfile8HPqsData(packet_type_id=SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_1_S__AVG, days_ago=timedelta(seconds=5443200.0), profile=(0, 1, 2, 3, 4, 5, 6, 7), point_factor=0.0) == SmpmUlDeviceEnergy16BProfile8HPqsData.parse(BufRef(case_serialized))  # noqa: E501
#     assert case_serialized == SmpmUlDeviceEnergy16BProfile8HPqsData.serialize(SmpmUlDeviceEnergy16BProfile8HPqsData(days_ago=timedelta(seconds=5443200.0), packet_type_id=SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_1_S__AVG, point_factor=0.0, profile=(0, 1, 2, 3, 4, 5, 6, 7)))  # noqa: E501
#     case_serialized = bytes.fromhex("b1cc0f8000086000042880010ef0ff00")
#     assert SmpmUlDeviceEnergy16BProfile8HPqsData(packet_type_id=SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_1_S__AVG, days_ago=timedelta(seconds=5443200.0), profile=(0, 1, 2, 3, 4, 5, 6, 7), point_factor=409.5) == SmpmUlDeviceEnergy16BProfile8HPqsData.parse(BufRef(case_serialized))  # noqa: E501
#     assert case_serialized == SmpmUlDeviceEnergy16BProfile8HPqsData.serialize(SmpmUlDeviceEnergy16BProfile8HPqsData(days_ago=timedelta(seconds=5443200.0), packet_type_id=SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_1_S__AVG, point_factor=409.5, profile=(0, 1, 2, 3, 4, 5, 6, 7)))  # noqa: E501
#     case_serialized = bytes.fromhex("9ccd0f8000086000042880010e000000")
#     assert SmpmUlDeviceEnergy16BProfile8HPqsData(packet_type_id=SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_3_Q__MAX__PHASE_C, days_ago=timedelta(seconds=5443200.0), profile=(0, 1, 2, 3, 4, 5, 6, 7), point_factor=0.0) == SmpmUlDeviceEnergy16BProfile8HPqsData.parse(BufRef(case_serialized))  # noqa: E501
#     assert case_serialized == SmpmUlDeviceEnergy16BProfile8HPqsData.serialize(SmpmUlDeviceEnergy16BProfile8HPqsData(days_ago=timedelta(seconds=5443200.0), packet_type_id=SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_3_Q__MAX__PHASE_C, point_factor=0.0, profile=(0, 1, 2, 3, 4, 5, 6, 7)))  # noqa: E501
#     case_serialized = bytes.fromhex("9ccd0f8000086000042880010ef0ff00")
#     assert SmpmUlDeviceEnergy16BProfile8HPqsData(packet_type_id=SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_3_Q__MAX__PHASE_C, days_ago=timedelta(seconds=5443200.0), profile=(0, 1, 2, 3, 4, 5, 6, 7), point_factor=409.5) == SmpmUlDeviceEnergy16BProfile8HPqsData.parse(BufRef(case_serialized))  # noqa: E501
#     assert case_serialized == SmpmUlDeviceEnergy16BProfile8HPqsData.serialize(SmpmUlDeviceEnergy16BProfile8HPqsData(days_ago=timedelta(seconds=5443200.0), packet_type_id=SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_3_Q__MAX__PHASE_C, point_factor=409.5, profile=(0, 1, 2, 3, 4, 5, 6, 7)))  # noqa: E501
